"use strict";

(function(window) {
    let ETSChat = {};

    String.prototype.padLeft = function(padString,length){
        var toReturn = String(this);
        while(toReturn.length < length){
            toReturn = padString + toReturn;
        }
        return toReturn;
    };

    class LocalStorage {
        constructor() {

        }

        write(key, value, days) {
            window.localStorage.setItem(key, value);
        }

        read(key) {
            return window.localStorage.getItem(key);
        }

        erase(key) {
            window.localStorage.removeItem(key);
        }

        clearAll() {
            window.localStorage.clear();
        }
    }

    class SessionStorage {
        constructor() {

        }

        write(key, value) {
            window.sessionStorage.setItem(key, value);
        }

        read(key) {
            return window.sessionStorage.getItem(key);
        }

        erase(key) {
            window.sessionStorage.removeItem(key);
        }

        clearAll() {
            window.sessionStorage.clear();
        }
    }

    class CookieStorage {
        constructor() {

        }

        write(key, value, inDays) {
            value = encodeURIComponent(value);
            const days = inDays || 30;
            var date = new Date();
            date.setTime(date.getTime() + (days * 24 * 60 * 60 * 1000));
            var expires = "; expires=" + date.toGMTString();
            document.cookie = key + "=" + value + expires + "; path=/";
        }

        read(key) {
            const nameEQ = key + "=";
            const ca = document.cookie.split(';');
            for (let i = 0, max = ca.length; i < max; ++i) {
                let c = ca[i];
                while (c.charAt(0) === ' ') c = c.substring(1, c.length);
                if (c.indexOf(nameEQ) === 0) {
                    return decodeURIComponent(c.substring(nameEQ.length, c.length));
                }
            }
            return null;
        }

        erase(key) {
            write(key, "", -1);
        }

        clearAll() {
            window.sessionStorage.clear();
        }
    }



    class SessionData {
        constructor(inDocument) {
            this._document = inDocument;
            this._entryUrl = "";
            this._refUrl = "";

            if (this._document.URL != null && this._document.URL !== '')
                this._entryUrl = document.URL;

            if (this._document.referrer != null && this._document.referrer !== '')
                this._refUrl = document.referrer;
        }
    }


    class Session {
        constructor(inFragmentId, inSocket, inDocument, inNavigator,
                    inScreen, inRetransmitTime, inHeadElement) {
            ETSChat.session = this;
            this._id = "";
            this._fragmentId = inFragmentId;
            this._socket = inSocket;
            this._document = inDocument;
            this._navigator = inNavigator;
            this._screen = inScreen;
            this._retransmitTime = inRetransmitTime;
            this._headElement = inHeadElement;
            this._cssStyle = '';
            this._cssTag = null;
            this._fragmentSettings = {};

            // see if we have a current session in our session storage
            const frag = ETSChat.cookieStorage.read(this._fragmentId);
            if (frag === null || frag === undefined) {
                // no session need to create one
                this.sendNewSessionRequest();
                console.log("no session found, creating new session");
            } else {
                const visitorID = JSON.parse(frag).visitorID;
                if(visitorID === null || visitorID == undefined) {
                    this.sendNewSessionRequest();
                } else {
                    // have a fragmentID in our cookie, now look for a session
                    // const sessionFrag = JSON.parse(ETSChat.sessionStorage.read(Session.SESSION_STORAGE_KEY));
                    let storedSessionData = Session.getStoredSessionData();
                    if(storedSessionData === null || storedSessionData === undefined) {
                        this.sendNewSessionRequest();
                    } else {
                        // have current session data in our session storage
                        // unpack it and proceed to 'page/new'
                        const sid = storedSessionData.sessionID;
                        if (sid !== null && sid !== '') { // New Page
                            this._id = sid;
                            this.sendNewPageRequest();
                            // initEventMessage.data.sessionType = 2; //New Page.
                            // initEventMessage.data.visitorID = vid;
                            // initEventMessage.data.sessionID = sid;
                            // eventMediator.mediateEvent(initEventMessage);
                        } else { // create new session
                            // initEventMessage.data.sessionType = 1; //New Session.
                            // initEventMessage.data.visitorID = vid;
                            // eventMediator.mediateEvent(initEventMessage);
                        }


                    }
                }
            }


            let storedSessionData = Session.getStoredSessionData();
            if(storedSessionData) {
                console.log("existing session found, using it");

                // load the visitor from the storage data
                // this._visitor = getVistoredSessionData
            } else {


            }
        }

        sendNewSessionRequest() {
            // create a vistor
            this._uaParser = new UAParser();
            this._uaParser.setUA(navigator.userAgent);
            this._uaParserResult = this._uaParser.getResult();
            this._visitor = new Visitor(document, navigator, screen);

            const sessionData = new SessionData(this._document);
            const newVisitorMessage = Visitor.createNewVisitorMessage(this._visitor,
                this._uaParserResult,
                sessionData,
                this._fragmentId);

            this._socket.open()
                .then((fulfilled, rejected) => {
                    if(fulfilled) {
                        // console.log('Socket.send() message:' , messageObject);
                        this._socket.sendRequest({message: newVisitorMessage})
                            .then((fulfilled, rejected) => {
                                if(fulfilled) {
                                    this.saveFragmentSettings(fulfilled.message.package);
                                    this._id = fulfilled.message.package.sessionID;
                                    this._visitor._id = fulfilled.message.package.visitorID;
                                    // ETSChat.messageDispatcher.handleMessage(fulfilled.message);
                                    console.log('Socket.send() wsp.sendRequest succeeded:', fulfilled.message);
                                    this.sendNewPageRequest();
                                } else if(rejected) {
                                    console.log('Socket.send() wsp.sendRequest failed for:', rejected.message);
                                }
                            });
                    } else if (rejected) {
                        console.log('Socket.send() open failed for:', rejected.message);
                    }
                });
        }

        sendNewPageRequest() {
            const newPageMessage = this.createNewPageMessage(this._document.entryUrl);
            this._socket.open()
                .then((fulfilled, rejected) => {
                    if(fulfilled) {
                        // console.log('Socket.send() message:' , messageObject);
                        this._socket.sendRequest({message: newPageMessage})
                            .then((fulfilled, rejected) => {
                                if(fulfilled) {
                                    // this.saveFragmentSettings(fulfilled.message.package);
                                    // ETSChat.messageDispatcher.handleMessage(fulfilled.message);
                                    this.handleNewPageResponse(fulfilled.message.package);
                                    console.log('Socket.send() wsp.sendRequest succeeded:', fulfilled.message);
                                } else if(rejected) {
                                    console.log('Socket.send() wsp.sendRequest failed for:', rejected.message);
                                }
                            });
                    } else if (rejected) {
                        console.log('Socket.send() open failed for:', rejected.message);
                    }
                });
        }

        handleNewPageResponse(response) {
            ETSChat.sessionStorage.write('uuid', response['uuid']);

            let sessionFragSettings = Session.getStoredSessionData();
            if(sessionFragSettings === null) {
                sessionFragSettings = {};
            }
            let isSMSButtonAllowed = response['isSMSButtonEnabled'] === '1';

            this._cssTag = this._document.createElement('style');
            this._cssTag.setAttribute('type', 'text/css');
            this._cssTag.media = 'all';

            if(sessionFragSettings.customCss === undefined) {
                sessionFragSettings.customCss = '';
            }
            if (sessionFragSettings.customCss !== null && sessionFragSettings.customCss !== undefined &&
                sessionFragSettings.customCss !== '') {
                let decodedCss = Base64.decode(sessionFragSettings.customCss);
                let concatCss = this._cssStyle + " " + decodedCss;

                if (this._cssTag.styleSheet)
                    this._cssTag.styleSheet.cssText = concatCss;
                else
                    this._cssTag.appendChild(this._document.createTextNode(concatCss));
            } else {
                if (this._cssTag.styleSheet)
                    this._cssTag.styleSheet.cssText = this._cssStyle;
                else
                    this._cssTag.appendChild(this._document.createTextNode(this._cssStyle));
            }

            //
            // this._headElement.appendChild(this._cssTag);

            const modulesEnabled = sessionFragSettings.modulesEnabled === '1';


            ETSChat.chatButtonController.handleChatEnabled(sessionFragSettings);
            ETSChat.chatWindowController.handleChatEnabled(sessionFragSettings);

            if(isSMSButtonAllowed) {
                ETSChat.EventMediator.mediateEvent(ETSChat.enableSMSButton());
            }

            //ETSChat.EventMediator.mediateEvent(ETSChat.pollMessagesMessage());
            //sessionRuleTimer.setEnabled(1);
        }

        showChatButton() {
            console.log("Session::showChatButton");
        }

        static get SESSION_STORAGE_KEY() { return 'etsSessionData'; }

        static getStoredSessionData() {
            const data = window.sessionStorage.getItem(Session.SESSION_STORAGE_KEY);
            if(data) {
                const parsedData = JSON.parse(data);
                return parsedData;
            }
            return null;
        }

        saveFragmentSettings(sessionResponse) {
            const cookiefragSettings = {
                visitorID: sessionResponse['visitorID']
            }
            const websiteBusinessHours = sessionResponse['businessHours'];
            const modules = sessionResponse['modulesEnabled'];
            const boundHTMLElements = sessionResponse['boundHTMLElements'];
            /*
             * Mandatory data writes.
             */

            const isSMSButtonAllowed = sessionResponse['isSMSButtonEnabled'];
            const sessionFragSettings = {
                sessionID: sessionResponse['sessionID'],
                websiteTimeZoneOffset: sessionResponse['timezoneOffset'],
                websiteDefaultThumbnail: sessionResponse['websiteDefaultPic'],
                companyID: sessionResponse['companyID'],
                customCss: "",
                boundHTMLElements: [],
                businessHours: [],
                modulesEnabled: modules,
            }

            if (websiteBusinessHours !== undefined) {
                for (let i = 0; i < websiteBusinessHours.length; i++)
                    sessionFragSettings.businessHours.push(JSON.parse(websiteBusinessHours[i]));
            }

            if (boundHTMLElements !== undefined) {
                for (let i = 0; i < boundHTMLElements.length; i++) {
                    sessionFragSettings.boundHTMLElements.push(boundHTMLElements[i]);
                }
            }

            //alert("session frag settings offsconstet " + sessionFragSettings["websiteTimeZoneOffset"]);
            const chat = sessionResponse['chat'];
            const usersInChat = sessionResponse['userIdsInChat'];
            const thumbnailsInChat = sessionResponse['userProfileThumbs'];
            const websiteDefaultThumbnail = sessionResponse['websiteDefaultPic'];

            //Agent name resurrection vars.
            const agentNameIdsInChat = sessionResponse['userDisplayAgentNameIds'];
            const agentFirstNames = sessionResponse['firstName'];
            const agentLastNames = sessionResponse['lastName'];
            const agentDisplayNames = sessionResponse['displayName'];

            // if ((modules) & 1) {
            if (true) {
                //Init load html Structure and css (in case there are no overiding options sent across).
                sessionFragSettings.chatButton = JSON.parse(sessionResponse['chatButton']);
                sessionFragSettings.chatInactiveButton = JSON.parse(sessionResponse['chatInactiveButton']);
                sessionFragSettings.chatWindow = JSON.parse(sessionResponse['chatWindow']);

                if ((modules >> 0x01) & 1)
                    sessionFragSettings.emailSettings = JSON.parse(sessionResponse['emailSettings']);

                if ((modules >> 0x02) & 1)
                    sessionFragSettings.preChatForm = JSON.parse(sessionResponse['preChatForm']);

                if ((modules >> 0x03) & 1)
                    sessionFragSettings.postChatForm = JSON.parse(sessionResponse['postChatForm']);
            }

            /*if ((modules >> 0x04) & 1)
                sessionFragSettings.css = inMessageObject['customCss'];*/

            if (sessionResponse.customCss !== undefined && sessionResponse.customCss !== "") {
                sessionFragSettings.customCss = sessionResponse.customCss;
            } else {
                sessionFragSettings.customCss = '';
            }

            //HMM THIS SHOULD BE LOADED AT BOOTSTRAP TIME?TODO!!!:::
            this._sessionFragSettings = sessionFragSettings;

            ETSChat.sessionStorage.write(Session.SESSION_STORAGE_KEY, JSON.stringify(sessionFragSettings));
            ETSChat.sessionStorage.write('uuid', sessionResponse['uuid']);

            ETSChat.cookieStorage.write(this._fragmentId, JSON.stringify(cookiefragSettings));
            ETSChat.sessionStorage.write("newVisitor", true);

        }

        createNewPageMessage(inSessionUrl) {
            let nRecoverableMsgs = -1;
            let chat = ETSChat.sessionStorage.read('chat');
            if(chat!== undefined && chat !== '' && chat!==null) {
                chat = JSON.parse(chat);

                //should only be null or undefined if this was an existing chat from before nRecoverableMessages was added
                //so we pass -1 to disable message replay for this page session
                if( typeof chat.nRecoverableMessages === 'undefined' ) {
                    chat.nRecoverableMessages = -1;
                }
                if( chat.nRecoverableMessages === null ) {
                    chat.nRecoverableMessages = -1;
                }
                nRecoverableMsgs = chat.nRecoverableMessages;
            }

            const storedSession = JSON.parse(ETSChat.sessionStorage.read(Session.SESSION_STORAGE_KEY));

            let messageObject = {
                oty 		: 'tco',
                mty 		: 'request',
                rty 		: 'page/new',
                package     : {
                    fragmentID 	     : this._fragmentId,
                    sessionID        : storedSession.sessionID,
                    visitorID        : storedSession.visitorID,
                    url              : this._document.entryUrl,
                    nRecoverableMessages : nRecoverableMsgs
                }
            };
            return messageObject;
        }

    }

    class Visitor {
        constructor(inDocument, inNavigator, inScreen) {
            this._id = "";
            this._document 	= inDocument;
            this._navigator 	= inNavigator;
            this._screen 		= inScreen;
            this._data = {
                // Browser defined
                language 				       : '',
                platform 				       : '',

                timeZone 				       : {
                    utOffset			           : ''
                },

                screenResolution 		       : {
                    totalWidthHeight		       : "",
                    availableWidthHeight	       : "",
                    colorDepth			       : "",
                    pixelDepth			       : ""
                },
            };

            if (this._navigator.language != null && this._navigator.language !== '')
                this._data.language = this._navigator.language;


            // getPlatform
            if (this._navigator.platform != null && this._navigator.platform !== '')
                this._data.platform = this._navigator.platform;

            // getTime Zone
            const date = new Date()
            const offset = -date.getTimezoneOffset() / 60;
            this._data.timeZone.utOffset += (offset < 0) ? '-' : '+';
            this._data.timeZone.utOffset += offset;

            // getScreenResolution
            if (this._screen.width != null && this._screen.height != null && this._screen.width !== '' && this._screen.height !== '')
                this._data.screenResolution.totalWidthHeight = this._screen.width + "/" + this._screen.height;

            if (this._screen.availWidth != null && this._screen.availHeight != null && this._screen.availWidth !== '' && this._screen.availHeight !== '')
                this._data.screenResolution.availableWidthHeight = this._screen.availWidth + "/" + this._screen.availHeight;

            if (this._screen.colorDepth != null && this._screen.colorDepth !== '')
                this._data.screenResolution.colorDepth = this._screen.colorDepth;

            if (this._screen.pixelDepth != null && this._screen.pixelDepth !== '')
                this._data.screenResolution.pixelDepth = this._screen.pixelDepth;
        }

        static createNewVisitorMessage(inVisitorData, inUAParserResult, inSessionData, inFragmentId) {
            const messageObject = {
                oty 		: 'tco',
                mty 		: 'request',
                rty 		: 'visitor/new',
                package : {
                    fragmentID 	    : inFragmentId,
                    language        : inVisitorData.language,
                    userAgent       : navigator.userAgent,
                    platform 	    : ETSUtils.checkVisitorIsOnMobile() ? 'mobile' : 'desktop',
                    timezoneOffset  : inVisitorData._data.timeZone.utOffset,
                    screenSize 	    : inVisitorData._data.screenResolution.totalWidthHeight,
                    browser	        : inUAParserResult.browser.name + ' ' + inUAParserResult.browser.version,
                    operatingSystem : inUAParserResult.os.name + ' ' + inUAParserResult.os.version,
                    device 	        : (inUAParserResult.device.vendor === undefined) ? '' : inUAParserResult.device.vendor + ' ' + inUAParserResult.device.type + ' ' + inUAParserResult.device.model,
                    entryUrl 	    : inSessionData.entryUrl,
                    referrerUrl     : inSessionData.refUrl
                }
            };
            return messageObject;
        }
    }

    class ChatButtonController {
        constructor(inView, inSocket, inFragmentId) {
            this._view = inView;
            this._socket = inSocket;
            this._fragmentId = inFragmentId;

            // this._eventTypes = [TC.chatEnabledRequestType(), TC.showChatButtonRequestType(),
            //     TC.chatFinishedRequestType(), TC.serverErrorRequestType(),
``            //     TC.inviteVisitorToChatRequestType(), TC.hideChatButtonRequestType()];
            //
            // this._requestTypes = ["website/updateFragNotify", "chat/reopened", "visInvite/visitorResponse", "chat/end", "chat/new"];

            this._isChatWindowExternal = 0;
            this._sessionId = "";

            this._pollBusinessHoursTimeout = 4000;
            // this._businessHoursTimer = new TC.Timer(pollBusinessHours, pollBusinessHoursTimeout, 0);
            this._businessHoursTimer = null;

            this._todaysDay = -1;
            this._visitorsDayAfterOffset = -1;
            this._visitorsDayAfterOffsetCompare = -1;

            this._startTime = -1;
            this._endTime = -1;

            this._parsedBusinessHours = -1;

            this._webServerOffset = -1;

            this._isBusinessHoursInit = 0;
            this._dayChecked = 0;
            this._justCheckDay = 0;

            //var disabledChatButtonActive = 0;
            //var disabledChatButtonShown = 0;
            this._isChatButtonOffline = false;

            this._state = 0;
            //variables for offline business hours form.
            this._convertedOpeningHours = {};
            this._convertedEndingHours  = {};
            this._stateArray = {};

            this._isServerError = 0;

            this._tcFragSettings = {};
            this._invited = 0;

            this._isChatEnabled = 0;

            this._inChat = 0;

            this._visitorInviteButtonNotHit = true;

            this._view.setChatBtnClickedCallback(ChatButtonController.handleChatBtnClicked);
            this._view.setDisabledChatBtnClickedCallback(this.handleDisabledChatBtnClicked);
            this._view.setLiveInviteClickedCallback(this.handleLiveChatInviteBtnClicked);
            this._view.setLiveInviteCloseBtnClickedCallback(this.handleLiveChatInviteCloseBtnClicked);

            // this._view.showChatButton();

        }


        //Private functions.
        //***************************************************************************
        //* Mediator Event Handling                                                 *
        //***************************************************************************

        static handleChatBtnClicked() {
            let chatButtonController = ETSChat.chatButtonController;
            if (chatButtonController._invited) {
                chatButtonController._invited = 0;
                chatButtonController._view.removeLiveChatInvite();
            }
            chatButtonController._inChat = ETSChat.sessionStorage.write('inChat', 1);
            const frag = Session.getStoredSessionData();
            const displayEmbeddedChatWindowOnMobile = frag.chatWindow.displayEmbeddedChatWindowOnMobile;
            //TODO need to rethink visitor mobile.
            if (!(chatButtonController._isChatWindowExternal) && !ETSUtils.checkVisitorIsOnMobile()) {
                // MGMG TODO -- implement chat window here
                // ETSChat.chatButtonController.handleChatButtonClicked();
                ETSChat.chatWindowController.handleChatButtonClicked();
                // TC.EventMediator.mediateEvent(TC.chatButtonClickedMessage());
                chatButtonController._view.hideChatButton();
                // this._businessHoursTimer.setEnabled(0);
                chatButtonController._inChat = 1;
            } else {
                if(displayEmbeddedChatWindowOnMobile === true) {
                    // TC.EventMediator.mediateEvent(TC.chatButtonClickedMessage());
                    // ETSChat.chatButtonController.handleChatButtonClicked();
                    ETSChat.chatWindowController.handleChatButtonClicked();
                    chatButtonController._view.hideChatButton();
                    // this._businessHoursTimer.setEnabled(0);
                    chatButtonController._inChat = 1;
                } else {
                    window.open("https://www.etschat.com/script/TonicChatNewWindow.html?wfid=" + chatButtonController._fragmentId
                        + "&sid=" + chatButtonController._sessionId + "&vid=" + frag.visitorID, "ETSChat, powered by Engage To Sell LLC", "height=600, width=800");
                }
            }
        }


        handleDisabledChatBtnClicked() {

            //Roman: do I want in this moment set button to invisible ? or wait for some server response...
            this._view.hideChatButton();
            // this._businessHoursTimer.setEnabled(0);
            convertBusinessHoursOpeningTimes();
        }


        handleLiveChatInviteBtnClicked() {
            this._visitorInviteButtonNotHit = false;
            this._invited = 0;
            this._view.removeLiveChatInvite();
            const frag = Session.getStoredSessionData();
            const displayEmbeddedChatWindowOnMobile = frag.chatWindow.displayEmbeddedChatWindowOnMobile;
            //TODO need to rethink visitor mobile.
            if (!(this._isChatWindowExternal) && !ETSUtils.checkVisitorIsOnMobile()) {
                // MGMG TODO fix this
                // this._socket.send(TC.createVisitorInviteAcceptedProtocol(inFragmentId));
                // TC.EventMediator.mediateEvent(TC.visitorInviteAcceptedMessage());
                this._view.hideChatButton();
                // this._businessHoursTimer.setEnabled(0);
                this._inChat = 1;
            } else {
                if(displayEmbeddedChatWindowOnMobile === true) {
                    // this._socket.send(TC.createVisitorInviteAcceptedProtocol(inFragmentId));
                    // TC.EventMediator.mediateEvent(TC.visitorInviteAcceptedMessage());
                    this._view.hideChatButton();
                    // this._businessHoursTimer.setEnabled(0);
                    this._inChat = 1;
                } else {
                    const chatId = chatWindowController.getChatId();
                    window.open("https://www.etschat.com/script/TonicChatNewWindow.html?wfid=" + this._fragmentId
                        + "&sid=" + this._sessionId + "&vid=" + frag.visitorID, "ETSChat, powered by Engage To Sell LLC", "height=600, width=800");
                }
            }
        }


        handleLiveChatInviteCloseBtnClicked() {
            // TODO Fix this
            // this._socket.send(TC.createVisitorInviteDeclinedProtocol(inFragmentId));
            this._invited = 0;
            // TC.EventMediator.mediateEvent(TC.chatEndMessage());

            this._view.showChatButton();
            this._isChatButtonOffline = false;
        }


        handleChatEnabled(fragmentSettings) {
            this._sessionId = fragmentSettings.sessionID;
            this._view.setUISettings(fragmentSettings.chatButton, fragmentSettings.chatInactiveButton);
            this._isChatWindowExternal = fragmentSettings.chatWindow.openExternalWindow;

            const bindingHTMLElements = fragmentSettings.boundHTMLElements;

            for(let i = 0; i < bindingHTMLElements.length; i++) {
                let element = document.getElementById(bindingHTMLElements[i]);
                if (element !== null && element !== undefined)
                    element.onclick = this.handleChatBtnClicked;
            }

            this._view.showChatButton();
        }

        handleVisitorInviteToChat() {
            this._view.renderChatInvite();
            //inView.hideChatButton();
        }

/*
        handleEvent(inEventMessage) {
            switch(inEventMessage.eventType) {
                case TC.chatEnabledRequestType():
                    this._isChatEnabled = 1;
                    handleChatEnabled(inEventMessage);
                    this._tcFragSettings = inEventMessage['data']['fragmentSettings'];
                    break;
                case TC.inviteVisitorToChatRequestType():
                    const message = TC.chatEnabledMessage();
                    if(isChatEnabled == 0) {
                        var frag = JSON.parse(TC.sessionStorage.read('tcFragSettings'));
                        if((frag.modulesEnabled & 1)) {
                            isChatEnabled = 1;
                        }
                    }
                    inChat = TC.sessionStorage.read('inChat');
                    if(isChatEnabled == 1 && inChat != 1) {
                        handleVisitorInviteToChat();
                        invited = 1;
                    }
                    break;
                case TC.showChatButtonRequestType():
                    pollBusinessHours();
                    businessHoursTimer.setEnabled(1);
                    break;
                case TC.chatFinishedRequestType():
                    if (!isServerError) {
                        if(isChatButtonOffline == false) {
                            inView.showChatButton();
                        } else {
                            inView.showDisabledChatButton();
                            businessHoursTimer.setEnabled(1);
                        }
                    }
                    break;
                case TC.serverErrorRequestType():
                    businessHoursTimer.setEnabled(0);
                    break;
                case TC.hideChatButtonRequestType():
                    inView.hideChatButton();
                    break;
                default:
                    throw new Error("Chat button controller attempted to handle invalid event type");
            }
        }
*/

        handleRequest(inRequestType, inMessageObject) {
            switch(inRequestType) {
                case "website/updateFragNotify": {
                    //TODO::TEdwards
                    //This is a hacky approach for the demonstration, need to rethink this post friday.
                    let diff = JSON.parse(inMessageObject.diff);
                    if (diff.chatButton !== undefined && diff.chatButton !== null && diff.chatButton !== "") {
                        if (diff.chatButton.bakedImage !== undefined && diff.chatButton.bakedImage !== null && diff.chatButton.bakedImage !== "")
                            this._view.changeChatButtonImage(diff.chatButton.bakedImage);
                        if(diff.chatButton.alignment !== undefined && diff.chatButton.alignment !== null && diff.chatButton.alignment !== "" && diff.chatButton.alignment !== 0)
                            this._view.setButtonAlignment(diff.chatButton.alignment);
                    }
                    break;
                }
                case "chat/reopened":
                    this._view.hideChatButton();
                    break;
                case "visInvite/visitorResponse":
                    if(inMessageObject.visitorInviteReaction === true && visitorInviteButtonNotHit === true) {
                        handleLiveChatInviteBtnClicked();
                    } else if(inMessageObject.visitorInviteReaction === false) {
                        this._view.removeLiveChatInvite();
                        this._view.showChatButton();
                        this._isChatButtonOffline = false;
                    }
                    break;
                case "chat/end":
                    this._visitorInviteButtonNotHit = true;
                    break;
                case "chat/new":
                    if (this._invited) {
                        this._invited = 0;
                        this._view.removeLiveChatInvite();
                    }
                    break;
                default: {break;}
            }
        }

        convertBusinessHoursOpeningTimes() {

/*            let vis                    = TC.getVisitorData(this.document, this.navigator, this.screen);
            var visitorsTimeZoneOffset = vis.timeZone.utOffset * 600 * 6000;

            for(var i = 0; i < 7; i++) {
                var busHours = tcFragSettings.businessHours[i];

                var    startTime    = busHours.availabilityFrom;
                var    endTime      = busHours.availabilityTo;
                var    state        = busHours.state;

                if(visitorsTimeZoneOffset < webServerOffset) {
                    startTime += webServerOffset - visitorsTimeZoneOffset;
                    endTime   += webServerOffset - visitorsTimeZoneOffset;
                } else {
                    if(webServerOffset - visitorsTimeZoneOffset > 0) {
                        startTime -= webServerOffset - visitorsTimeZoneOffset;
                        endTime   -= webServerOffset - visitorsTimeZoneOffset;
                    } else {
                        startTime += webServerOffset - visitorsTimeZoneOffset;
                        endTime   += webServerOffset - visitorsTimeZoneOffset;
                    }
                }

                var timeStartHour =  new Date(startTime).getHours();
                var timeStartMin  =  new Date(startTime).getMinutes();
                var timeStartSec  =  new Date(startTime).getSeconds();

                if(timeStartHour < 10) {
                    timeStartHour = 0 + "" + timeStartHour;
                }

                if(timeStartMin < 10) {
                    timeStartMin = 0 + "" + timeStartMin;
                }


                var startFullTime = timeStartHour + ":" + timeStartMin; /!*+ ":" + timeStartSec;*!/

                var timeEndHour =  new Date(endTime).getHours();
                var timeEndMin  =  new Date(endTime).getMinutes();
                //var timeEndSec  =  new Date(endTime).getSeconds();

                if(timeEndHour < 10) {
                    timeEndHour = 0 + "" + timeEndHour;
                }

                if(timeEndMin < 10) {
                    timeEndMin = 0 + "" + timeEndMin;
                }


                var endFullTime = timeEndHour + ":" + timeEndMin; /!*+ ":" + timeEndSec;*!/


                convertedOpeningHours[i] = startFullTime;
                convertedEndingHours[i]  = endFullTime;
                stateArray[i]            = state;
            }
            //Call an event containing the converted opening hours array and the ending hours array.
            var message = TC.openingHoursViewMessage();
            message.data.convertedOpeningHours = convertedOpeningHours;
            message.data.convertedEndingHours  = convertedEndingHours;
            message.data.stateArray            = stateArray;
            message.data.companyEmail          = tcFragSettings.chatInactiveButton.textEmailOffer;
            message.data.emailCompanyQuestion  = tcFragSettings.chatInactiveButton.emailOffer;
            message.data.displayBusinessHours  = tcFragSettings.chatInactiveButton.displayBusinessHoursWhenOfflineCheck;

            TC.EventMediator.mediateEvent(message);*/
        }

        pollBusinessHours() {

/*            //var state = lastState;
            // frm Dave: setting state to zero each time without calling init makes it invalid on the next polling (it will always be zero)
            // rather than being determined by parsing .. so for now I left it unitialized so it will be the same state as last.

            var msecsSinceEpoch = new Date().getTime();


            //converts the offsets to mSecs.
            webServerOffset = tcFragSettings['websiteTimeZoneOffset'] * 1000;
            // server sends over seconds, so ms is simply times 1000
            //webServerOffset = tcFragSettings['websiteTimeZoneOffset'] * 1000;

            var date                   = new Date();
            var visitorsTimeZoneOffset = -date.getTimezoneOffset() / 60 * 600 * 6000;


            if(visitorsTimeZoneOffset < webServerOffset) {
                msecsSinceEpoch += webServerOffset - visitorsTimeZoneOffset;
            } else {
                if(webServerOffset - visitorsTimeZoneOffset > 0) {
                    msecsSinceEpoch -= webServerOffset - visitorsTimeZoneOffset;
                } else {
                    msecsSinceEpoch += webServerOffset - visitorsTimeZoneOffset;
                }
            }

            //sets the current day of the vistor, converted to the servers offset to ensure the correct day is used.
            visitorsDayAfterOffset = new Date(msecsSinceEpoch).getDay();

            //updates the the compare variable.
            if(dayChecked == 0) {
                visitorsDayAfterOffsetCompare = visitorsDayAfterOffset;
                dayChecked = 1;
            }
            //if the day changes make sure it inits to grab the new business hours times for the new day.
            if(visitorsDayAfterOffset != visitorsDayAfterOffsetCompare) {
                isBusinessHoursInit = 0;
                justCheckDay = 0;
                dayChecked = 0;
            }

            todaysDay = visitorsDayAfterOffset;

            //Converts to the clients day setup.
            if(todaysDay == 0)
                todaysDay += 1;
            else
                todaysDay -= 1;

            if(justCheckDay != 1) {

                if(isBusinessHoursInit !== 1) {

                    /!* displayWhenInActive is not being saved into tcFragSettings *!/
                    //chatButtonSettings = tcFragSettings.chatButton;
                    //disabledChatButtonActive = chatButtonSettings['displayWhenInActive'];

                    //grab the business hours objects for the given day.
                    for(var i = 0; i < 7; i++) {
                        parsedBusinessHours = tcFragSettings.businessHours[i]
                        if(i == todaysDay) {
                            state = parsedBusinessHours['state'];
                            startTime = parsedBusinessHours['availabilityFrom'];
                            endTime = parsedBusinessHours['availabilityTo'];
                            break;
                        }
                    }

                    isBusinessHoursInit = 1;
                }

                //checks to see if it needs to care about checking the times, or if chat is set to enabled by default.
                var businessHoursDisabled = 0;
                var businessHoursEnabledForSetTime = 1; // real times e=nabled all day = 2, hours disabled = 0
                var businessHoursEnabledAllDay = 2;

                if (state === businessHoursDisabled) {
                    inView.showDisabledChatButton();
                    isChatButtonOffline = true;
                    isChatEnabled = 0;
                    //disabledChatButtonShown = 1;
                    justCheckDay = 1;
                } else if (state === businessHoursEnabledAllDay) {
                    inView.showChatButton();
                    isChatButtonOffline = false;
                    justCheckDay = 1;
                } else {

                    var visitorsCurrentTime = new Date(msecsSinceEpoch);

                    var startOfDay = new Date(msecsSinceEpoch);
                    var visitorsCurrentTime = new Date(startOfDay);
                    var visitorsCurrentTimeFromStartOfDay = visitorsCurrentTime - startOfDay.setHours(0,0,0,0);

                    var outOfBusinessHours = 1;
                    //check if its a day or night shift.
                    var isNightShift = 0;
                    if(startTime > endTime) {
                        isNightShift = 1;
                    }

                    if(isNightShift == 1) {
                        if(visitorsCurrentTimeFromStartOfDay < startTime && visitorsCurrentTimeFromStartOfDay > endTime) {
                            outOfBusinessHours = 1;
                        } else {
                            outOfBusinessHours = 0;
                        }
                    } else if(isNightShift == 0) {
                        if(visitorsCurrentTimeFromStartOfDay > startTime && visitorsCurrentTimeFromStartOfDay < endTime || startTime == 0 && endTime == 0) {
                            outOfBusinessHours = 0;
                        }
                    }

                    if(outOfBusinessHours != 1) {
                        inView.showChatButton();
                        isChatButtonOffline = false;
                    } else {
                        inView.showDisabledChatButton();
                        isChatButtonOffline = true;
                    }
                }
            }
            //lastState = state;*/
        }

/*
        getEventTypes() {
            return this._eventTypes;
        }

        getRequestTypes() {
            return this._requestTypes;
        }

*/
    }


    var chatButtonClickedCallback           = null;
    var disabledChatButtonClickedCallback   = null;
    var liveInviteClickedCallback           = null;
    var liveInviteClosedBtnClickedCallback  = null;


    class ChatButtonView {
        constructor(pageBodyElement, alignmentManager) {
            this._bodyElement            = pageBodyElement;
            this._alignmentManager       = alignmentManager;
            this._buttonPlatform 		= null;
            this._chatButtonDiv          = null;
            this._buttonVisibilityDiv    = null;
            this._liveChatInviteDiv      = null;
            this._chatButtonSettings     = null;
            this._buttonGroupDiv       = null;
            this._liveChatBtnCloseDiv    = null;

            this._currentAlignmentNumber = null;

            const sessionData = Session.getStoredSessionData();
            // var obj = JSON.parse(TC.sessionStorage.read("tcFragSettings"));

            const ChatButtonType     = {
                kHyperLink:           1,
                kButtonImg:           2
            };

        }

        handleClickEvent(e) {
            if (chatButtonClickedCallback !== null)
                chatButtonClickedCallback();
        }

        handleClickEventForDisabledChatButton(e) {
            if(disabledChatButtonClickedCallback !== null)
                disabledChatButtonClickedCallback();
        }

        handleLiveInviteClicked(e) {
            if (liveInviteClickedCallback !== null)
                liveInviteClickedCallback();
        }

        handleLiveInviteCloseBtnClicked(e) {
            if(liveInviteClosedBtnClickedCallback !== null)
                handleHideChatInviteButton();
            liveInviteClosedBtnClickedCallback();
        }

        handleShowChatButton() {
            //TC.setStyleDisplay(mChatButtonDiv, 1);

            this._buttonVisibilityDiv.className = "tcButtonVisible";
            this._chatButtonDiv.className = "tcButtonStateOnline";
            this._chatButtonDiv.onclick = this.handleClickEvent;
        }

        handleShowDisabledChatButton() {
            //TC.setStyleDisplay(mDisabledChatButtonDiv, 1);
            this._buttonVisibilityDiv.className = "tcButtonVisible";
            this._chatButtonDiv.className = "tcButtonStateOffline";
            this._chatButtonDiv.onclick = this.handleClickEventForDisabledChatButton;
        }

        handleHideChatButton() {
            //TC.setStyleDisplay(mChatButtonDiv, 0);
            this._buttonVisibilityDiv.className = "tcButtonInvisible";
            this._chatButtonDiv.onclick = null;
        }

        handleHideChatInviteButton() {
            ETSUtils.setStyleDisplay(mLiveChatInviteDiv, 0);
            this._liveChatInviteDiv.onclick = null;
        }

        //public methods.
        setChatBtnClickedCallback(inCallback) {
            chatButtonClickedCallback = inCallback;
        }

        setDisabledChatBtnClickedCallback(inCallback) {
            disabledChatButtonClickedCallback = inCallback;
        }

        setLiveInviteClickedCallback(inCallback) {
            liveInviteClickedCallback = inCallback;
        }

        setLiveInviteCloseBtnClickedCallback(inCallback) {
            liveInviteClosedBtnClickedCallback = inCallback;
        }

        changeChatButtonImage(inImageData) {
            const btnImg = mChatButtonDiv.firstChild;
            btnImg.innerHTML = "<img src ='data:image/" + "png" + ";base64," + inImageData + "'></img>";
            const fragSettings = Session.getStoredSessionData();
            fragSettings.chatButton.bakedImage = inImageData;
            fragSettings.chatButton.image = inImageData;
            ETSChat.sessionStorage.write(Session.SESSION_STORAGE_KEY, JSON.stringify(fragSettings));
        }

        renderChatInvite() {
            if (this._liveChatInviteDiv === null) {
                this._liveChatInviteDiv = document.createElement('div');
                //this._liveChatInviteDiv.id = "tcLiveChatInvite";

                if(TC.checkVisitorIsOnMobile()) {
                    this._liveChatInviteDiv.className = "tcButtonMobile";
                } else {
                    this._liveChatInviteDiv.className = "tcButtonDesktop";
                }

                const chatButtonState = document.createElement('div');
                chatButtonState.className = "tcButtonStateInvitation";

                this._liveChatInviteDiv.appendChild(chatButtonState);

                const kButtonId                      = "tcButtonLiveChat";
                const kAnimationClass                = "tcButtonAnimation";
                const kButtonTextArea                = "tcButtonTextArea";
                const kButtonText                    = "tcButtonText";
                const kButtonCloseIcon               = "tcButtonCloseIcon";

                const chatButton                       = document.createElement('div');
                chatButton.id                        = kButtonId;

                const chatButtonAnimation              = document.createElement('div');
                chatButtonAnimation.className        = kAnimationClass;

                const chatButtonTextArea               = document.createElement('div');
                chatButtonTextArea.className         = kButtonTextArea;

                const chatButtonText                   = document.createElement('div');
                chatButtonText.className             = kButtonText;

                this._liveChatBtnCloseDiv                 = document.createElement('div');
                //this._liveChatBtnCloseDiv.id              = "tcChatInviteCloseButton";
                this._liveChatBtnCloseDiv.className       = kButtonCloseIcon;
                this._liveChatBtnCloseDiv.onclick         = handleLiveInviteCloseBtnClicked;

                chatButtonTextArea.appendChild(chatButtonText);
                chatButtonTextArea.appendChild(this._liveChatBtnCloseDiv);
                chatButtonAnimation.appendChild(chatButtonTextArea);
                chatButton.appendChild(chatButtonAnimation);

                /*
                var chatButtonImg = document.createElement('div');

                chatButtonImg.innerHTML
                            = document.location.protocol == "https:" ?
                            "<img src ='https://www.etschat.com/script/lhm_chatbuttonbig.png' />"
                            :"<img src ='http://www.etschat.com/script/lhm_chatbuttonbig.png' />";

                chatButtonAnimation.insertBefore(chatButtonImg, chatButtonTextArea);
                */

                chatButtonState.appendChild(chatButton);
                this._bodyElement.appendChild(this._liveChatInviteDiv);
            }
            this._liveChatInviteDiv.onclick = liveInviteClickedCallback;
            ETSChat.setStyleDisplay(this._liveChatInviteDiv, 1);
        }

        removeLiveChatInvite() {
            this._liveChatInviteDiv.onclick = null;
            ETSChat.setStyleDisplay(this._liveChatInviteDiv, 0);
        }

        setButtonAlignment(alignment) {
            let alignmentNumber;

            if(alignment === undefined) {
                alignmentNumber = this._chatButtonSettings.alignment;
            } else {
                alignmentNumber = alignment;
            }

            if(this._currentAlignmentNumber) {
                ETSChat.alignmentManager.addWidget(this._buttonPlatform, alignmentNumber, this._currentAlignmentNumber);
            } else {
                ETSChat.alignmentManager.addWidget(this._buttonPlatform, alignmentNumber, 0);
            }

            this._currentAlignmentNumber = alignmentNumber;

            const fragSettings = Session.getStoredSessionData();
            fragSettings.chatButton.alignment = alignmentNumber;
            ETSChat.sessionStorage.write(Session.SESSION_STORAGE_KEY, JSON.stringify(fragSettings));
        }

        setUISettings(inChatButtonSettings, inInactiveChatButtonSettings) {

            //TODO::Turn this whole object and the controller supporting it, into a object to represent more than one button.
            // hence the button group div now.
            const kButtonDesktop                   = "tcButtonDesktop";
            const kButtonMobile                    = "tcButtonMobile";

            //var kButtonGroup                     = "tcButtonsGroup";
            const kBtnVisibleState                = "tcButtonVisible";
            const kOnlineStateClass                = "tcButtonStateOnline";
            const kButtonId                        = "tcButtonLiveChat";
            const kAnimationClass                  = "tcButtonAnimation";
            const kButtonTextArea                  = "tcButtonTextArea";
            const kButtonText                      = "tcButtonText";

            this._chatButtonSettings               = inChatButtonSettings;
            this._inactiveChatButtonSettings       = inInactiveChatButtonSettings;

            //mButtonGroupDiv                      = document.createElement('div');
            //mButtonGroupDiv.className            = kButtonGroup;

            this._buttonPlatform	               = document.createElement('div');

            if(ETSUtils.checkVisitorIsOnMobile()) {
                this._buttonPlatform.className = kButtonMobile;
            } else {
                this._buttonPlatform.className = kButtonDesktop;
            }

            //mButtonGroupDiv.appendChild(buttonPlatform);

            this._buttonVisibilityDiv               = document.createElement('div');
            this._buttonVisibilityDiv.className      = kBtnVisibleState;

            const chatButton                       = document.createElement('div');
            chatButton.id                        = kButtonId;

            const chatButtonAnimation              = document.createElement('div');
            chatButtonAnimation.className        = kAnimationClass;

            const chatButtonTextArea               = document.createElement('div');
            chatButtonTextArea.className         = kButtonTextArea;

            const chatButtonText                   = document.createElement('div');
            chatButtonText.className             = kButtonText;

            chatButtonTextArea.appendChild(chatButtonText);
            chatButtonAnimation.appendChild(chatButtonTextArea);
            chatButton.appendChild(chatButtonAnimation);

            const chatButtonImg = document.createElement('div');
            if (this._chatButtonSettings.image === null || this._chatButtonSettings.image === "") {
                chatButtonImg.innerHTML
                    = document.location.protocol == "https:"
                    ? "<img src ='https://www.etschat.com/script/ChatBtnEnabled.png' />"
                    : "<img src ='http://www.etschat.com/script/ChatBtnEnabled.png' />";
            } else {
                chatButtonImg.innerHTML
                    = "<img src ='data:image/png;base64,"
                    + this._chatButtonSettings.bakedImage + "'></img>";
            }

            this._chatButtonDiv                       = document.createElement('div');
            this._chatButtonDiv.className             = kOnlineStateClass;
            this._chatButtonDiv.appendChild(chatButton);
            chatButtonAnimation.insertBefore(chatButtonImg, chatButtonTextArea);
            //mButtonGroupDiv.appendChild(mChatButtonDiv);
            this._buttonVisibilityDiv.appendChild(this._chatButtonDiv);
            this._buttonPlatform.appendChild(this._buttonVisibilityDiv);
            // ETSUtils.setStyleDisplay(this._chatButtonDiv, 0);
            // ETSUtils.setStyleDisplay(this._disabledChatButtonDiv, 0);
            //inBodyElement.appendChild(mButtonGroupDiv);

            this.setButtonAlignment();
        }

        showChatButton() {
            this.handleShowChatButton();
        }

        showDisabledChatButton() {
            this.handleShowDisabledChatButton();
        }

        hideChatButton() {
            this.handleHideChatButton();
        }
    }


    class ChatMessages {
        static get chatPollBackOffMessage()         { return 'ChatPollBackOff'; }
        static get chatButtonClickedRequestType()   { return 'chatButtonClicked'; }
        static get chatEnabledRequestType()         { return 'chatEnabled'; }
        static get chatEndRequestType()             { return 'chatEnd'; }
        static get chatFinishedRequestType()        { return 'chatFinished'; }
        static get chatNewWindowInitRequestType()   { return 'chatNewWindowInit'; }
        static get chatResurrectedRequestType()     { return 'chatResurrected'; }
        static get chatStartedRequestType()         { return 'chatStarted'; }
        static get checkBusinessHoursRequestType()  { return 'checkBusinessHours'; }
        static get enableSMSButtonRequestType()     { return 'enableSMSButton'; }
        static get hideChatButtonRequestType()      { return 'hideChatButton'; }
        static get inviteVisitorToChatRequestType() { return 'inviteVisitorToChat'; }
        static get liveInviteAcceptedRequestType()  { return 'liveInviteAccepted'; }
        static get liveInviteInitRequestType()      { return 'liveChatInviteInit'; }
        static get openingHoursViewRequestType()    { return 'openingHoursView'; }
        static get pollMessagesRequestType()        { return 'pollMessages'; }
        static get serverErrorRequestType()         { return 'serverError'; }
        static get sessionInitRequestType()         { return 'sessionInit'; }
        static get showChatButtonRequestType()      { return 'showChatButton'; }
        static get smsButtonClickedRequestType()    { return 'smsButtonClicked'; }
        static get smsChatFinishedRequestType()     { return 'smsChatFinished'; }
        static get startChatRequestType()           { return 'startChat'; }
        static get stopBusinessHoursTimerRequestType() { return 'stopBusinessHoursTimer'; }
        static get visitorInviteAcceptedRequestType() { return 'visitorInviteAccepted'; }

    }

    var ChatMessageType = {
        kText                   : 0x01,
        kLiveInviteMessage      : 0x03,
        kJoinedChat             : 0x04,
        kVisitorMessage         : 0x05,
        kLeftChat               : 0x08,
    };

    var State = {
        kNone                   : 0x00,
        kInChat                 : 0x01,
        kInPreChat              : 0x02,
        kInPostChat             : 0x04,
        kInExportForm           : 0x08,
        kEnded                  : 0x10,
        kVisitorInvitePending   : 0x20,
        kInOfflineForm          : 0x40
    }


    class ChatWindowController {

        constructor(inSocket, inSocketRetransmitTime, inFragmentId, inChatFormView) {
            if (!typeof inSocket === WebSocketAsPromised)
                throw new Error("Controller has been instantiated with an invalid socket object ('WebSocketAsPromised').");

            this._wsp = inSocket;
            this._socketRetransmitTime = inSocketRetransmitTime;
            this._fragmentId = inFragmentId;
            this._chatFormView =  inChatFormView;

            this._chatState = State.kNone;
            this._previousChatState = State.kNone;

            this._chat = {};
            this._chatInvite = {};
            this._chatId = "";
            this._chatInviteId = "";
            this._companyId = "";
            this._sessionId = "";
            this._websiteDefaultThumbnail = "";
            this._thumbnailsAllowed = 0;

            //Agent display name vars.
            this._agentDisplayFirstNamesMap = [];
            this._agentDisplayLastNamesMap = [];

            this._agentDisplayFullNamesMap = [];

            this._agentDisplayFirstName = "";
            this._agentDisplayLastName = "";
            this._agentDisplayFirstNameKey = "";
            this._agentDisplayLastNameKey = "";

            this._agentDisplayName = "";
            this._agentFullDisplayName = "";

            //Fragment settings data.
            this._modulesEnabled = 0x00;
            this._chatWindowSettings = null;
            this._preChatFormSettings = null;
            this._postChatFormSettings = null;
            this._systemMessages = 1;

            this._chatJoined = false;

            this._postChatTimer = new ETSTimer(this.onChatFinished, 0, 0);
            this._agentThumbnailsMap = [];
            this._minimised = 0;

            this._chatButtonNotClicked = true;

            this._chatMessageCount = 0;

            this._chatFormView.setSendButtonClickedCallback(ChatWindowController.handleSendMessageClicked);
            this._chatFormView.setEndButtonClickedCallback(ChatWindowController.handleEndChatClicked);
            this._chatFormView.setUnpinButtonClickedCallback(this.handleUnpinClicked);
            this._chatFormView.setPrintButtonClickedCallback(this.handlePrintClicked);
            this._chatFormView.setExportButtonClickedCallback(this.handleExportClicked);
            this._chatFormView.setVisitorKeyPressCallback(this.handleVisitorTypingKeyPress);
            this._chatFormView.setCloseChatButtonClickedCallback(ChatWindowController.handleCloseChatClicked);
            this._chatFormView.setMinimisedClickedCallback(this.handleMinimiseClicked);
            this._chatFormView.setExpandedClickedCallback(this.handleExpandedClicked);

        }

/*
        var mEventTypes = [TC.chatEnabledRequestType(), TC.chatButtonClickedRequestType(),
            TC.chatResurrectedRequestType(), TC.serverErrorRequestType(),
            TC.visitorInviteAcceptedRequestType(), TC.openingHoursViewRequestType(),
            TC.chatEnabledRequestType(), TC.chatEndRequestType(), TC.showChatWindowRequestType()
        ];

        var mRequestTypes = ["chat/agentTyping", "chat/message", "chat/end", "chat/new", "chat/join", "chat/export", "chat/prechat", "chat/preChatFilled", "chat/postchat", "chat/postChatFilled",
            "chat/reopened", "visInvite/check", "chat/offlineQuestion", "chat/closed",
            "visInvite/visitorResponse", "chat/agentStoppedTyping", "chat/resync"
        ];
*/




        handleBuildOpeningHoursForm (inMessageObject) {
            let startTimes = inMessageObject.data.convertedOpeningHours;
            let endTimes = inMessageObject.data.convertedEndingHours;
            let state = inMessageObject.data.stateArray;
            let companyEmail = inMessageObject.data.comletpanyEmail;
            let emailAllowed = inMessageObject.data.emailCompanyQuestion;
            let displayBusinessHours = inMessageObject.data.displayBusinessHours;

            inChatFormView.buildOutOfHoursForm(startTimes, endTimes, state, companyEmail, emailAllowed, displayBusinessHours);
            this._chatState = State.kInOfflineForm;
        }

        handleSetAgentDisplayName(displayName, agentDisplayFirstName, agentDisplayLastName, userId) {


            if (displayName != "" && displayName != undefined) {
                this._agentDisplayFullNameKey = userId + "ab";
                this._agentDisplayFullNamesMap[agentDisplayFullNameKey] = displayName;
                ETSChat.sessionStorage.write(agentDisplayFullNameKey, displayName);

            } else {
                this._agentDisplayFirstNameKey = userId + "a";
                this._agentDisplayLastNameKey = userId + "b";

                this._agentDisplayFirstName = this._agentDisplayFirstName[0].toUpperCase() + agentDisplayFirstName.slice(1);
                this._agentDisplayFirstNamesMap[this._agentDisplayFirstNameKey] = this._agentDisplayFirstName;
                ETSChat.sessionStorage.write(this._agentDisplayFirstNameKey, this._agentDisplayFirstName);

                this._agentDisplayLastName = this._agentDisplayLastName[0].toUpperCase() + this._agentDisplayLastName.slice(1);
                this._agentDisplayLastNamesMap[this._agentDisplayLastNameKey] = this._agentDisplayLastName;
                ETSChat.sessionStorage.write(this._agentDisplayLastNameKey, this._agentDisplayLastName);
            }
        }

        handleSetAgentThumbnail(inMessageObject) {
            this._agentThumbnailsMap[inMessageObject.userID] = inMessageObject.profPicture;
            ETSChat.sessionStorage.write(inMessageObject.userID, inMessageObject.profPicture);
        }

        onChatFinished() {
            this._chatState = State.kNone;
            this._postChatTimer.setEnabled(0);
            this._chatFormView.clearChatWindow();
            this._chatFormView.disableChat();
            this._chatFormView.hideChatWindow();
            // MGMG TODO -- fix this
            // TC.EventMediator.mediateEvent(TC.chatFinishedMessage());
            ETSChat.chatButtonView.showChatButton();
        }

        handleAgentTyping() {
            this._chatFormView.setNotification("Agent is typing...");
        }

        handleAgentStoppedTyping() {
            this._chatFormView.setNotification("");
        }

        static createChatMessage(inChatId, inCompanyId, inMsg, inFragmentId) {
            var messageObject = {
                oty: 'tco',
                mty: 'message',
                rty: 'chat/visitorMessage',
                package: {
                    sessionID: ETSChat.session._id,
                    fragmentID: inFragmentId,
                    uuid: ETSChat.sessionStorage.read('uuid'),
                    chatID: inChatId,
                    companyID: inCompanyId,
                    message: inMsg
                }
            }
            return messageObject;
        }

        sendNewChatMessageMessage(newChatMessageMessage) {
            this._wsp.open()
                .then(() => {
                    this._chatFormView._chatInput.value = "";
                    this._chatFormView._chatInput.focus();
                    this._wsp.sendRequest({message: newChatMessageMessage})
                        .then( fulfill => {
                            console.log('fulfill request:', fulfill.message);
                            this.addMessage(fulfill.message);
                        })
                        .catch((error) => {
                            console.log('catch error:', error);
                        });
                        // .finally(() => {
                        //     console.log('finally');
                        // });
                });
        }

        static handleSendMessageClicked(message) {
            let chatWindowController = ETSChat.chatWindowController;
            if (chatWindowController._chatState & 1) {
                if (message !== "") {
                    // MGMG TODO - implement with WebsocketAsPromised
                    // this._socket.send(TC.createChatMessageProtocol(mChatId, mCompanyId, message, inFragmentId));
                    let newChatMessageMessage = ChatWindowController.createChatMessage(chatWindowController._chatId,
                        "", message, chatWindowController._fragmentId);
                    chatWindowController.sendNewChatMessageMessage(newChatMessageMessage);
                }
            } else if (chatWindowController._chatState & State.kInPreChat) {

                const preChatForm = chatWindowController._chatFormView.getInChatForm("tcPreChatForm");

                const preChatObject = {
                    email: "",
                    firstName: "",
                    lastName: "",
                    initialQuestion: "",
                    questions: []
                };

                if (preChatForm.childNodes.length) {
                    for (let i = 0; i < preChatForm.childNodes.length; i++) {
                        let chatQuestion = preChatForm.childNodes[i].childNodes[0];
                        let name = "";
                        let value = "";
                        switch (chatQuestion.className) {
                            case "tcRadioQuestion":
                                name = chatQuestion.innerHTML;
                                let inputNode = chatQuestion.nextSibling;
                                value = [];
                                while (inputNode !== null) {
                                    if (inputNode.childNodes[0].checked) {
                                        value.push(inputNode.childNodes[0].nextSibling.innerHTML);
                                        inputNode = null;
                                    } else {
                                        inputNode = inputNode.nextSibling;
                                    }
                                }
                                break;
                            case "tcCheckQuestion":
                                name = chatQuestion.innerHTML;
                                inputNode = chatQuestion.nextSibling;
                                value = [];
                                while (inputNode !== null) {
                                    if (inputNode.childNodes[0].checked) {
                                        value.push(inputNode.childNodes[0].nextSibling.innerHTML);
                                    }
                                    inputNode = inputNode.nextSibling;
                                }
                                break;
                            case "tcEditQuestion":
                                let cId = chatQuestion.id;
                                if (cId === "tcQEmail") {
                                    preChatObject.email = chatQuestion.nextSibling.childNodes[0].value; //.innerHTML;
                                } else if (cId === "tcQNameFirst") {
                                    preChatObject.firstName = chatQuestion.nextSibling.childNodes[0].value; //.innerHTML;
                                } else if (cId === "tcQNameLast") {
                                    preChatObject.lastName = chatQuestion.nextSibling.childNodes[0].value; //.innerHTML;
                                } else if (cId === "tcQInitialQuestion") {
                                    preChatObject.initialQuestion = chatQuestion.nextSibling.childNodes[0].value; //.innerHTML;
                                } else {
                                    value = [];
                                    name = chatQuestion.innerHTML;
                                    value.push(chatQuestion.nextSibling.childNodes[0].value /*innerHTML*/ );
                                }
                                break;
                            default:
                                throw new Error("Chat Window Controller failed to render chat form data, invalid class name");
                                break;
                        }
                        if (value !== "") {
                            preChatObject.questions.push({
                                question: name,
                                answers: value
                            });
                        }

                    }
                }

                /*if (preChatForm === null)
                    throw new Error("Failed to get prechat form unable to gather pre chat data");
                else
                    nenderChatFormData(preChatForm);*/

                chatWindowController.handlePreChatFilled(JSON.stringify(preChatObject));

            } else if (chatWindowController._chatState & State.kInPostChat) {

                if (chatWindowController._postChatFormSettings.postChatAutoClose)
                    chatWindowController._postChatTimer.setEnabled(0);

                let postChatForm = chatWindowController._chatFormView.removeInChatForm("tcPostChatForm");
                let postChatObject = {
                    rating: "",
                    feedback: "",
                    questions: []
                };

                if (postChatForm.childNodes.length) {
                    for (let i = 0; i < postChatForm.childNodes.length; i++) {
                        let chatQuestion = postChatForm.childNodes[i].childNodes[0];
                        let name = "";
                        let value = "";
                        switch (chatQuestion.className) {
                            case "tcRadioQuestion":
                                let cId = chatQuestion.id;
                                name = chatQuestion.innerHTML;
                                let inputNode = chatQuestion.nextSibling;

                                if (cId === "tcQChatRating") {
                                    while (inputNode !== null) {
                                        if (inputNode.childNodes[0].checked) {
                                            value = inputNode.childNodes[0].nextSibling.innerHTML;
                                            inputNode = null;
                                        } else {
                                            inputNode = inputNode.nextSibling;
                                        }
                                    }
                                    postChatObject.rating = value;
                                    value = "";
                                } else {
                                    value = [];
                                    while (inputNode !== null) {
                                        if (inputNode.childNodes[0].checked) {
                                            value.push(inputNode.childNodes[0].nextSibling.innerHTML);
                                            inputNode = null;
                                        } else {
                                            inputNode = inputNode.nextSibling;
                                        }
                                    }
                                }

                                break;
                            case "tcCheckQuestion":
                                name = chatQuestion.innerHTML;
                                inputNode = chatQuestion.nextSibling;
                                value = [];
                                while (inputNode !== null) {
                                    if (inputNode.childNodes[0].checked) {
                                        value.push(inputNode.childNodes[0].nextSibling.innerHTML);
                                    }
                                    inputNode = inputNode.nextSibling;
                                }
                                break;
                            case "tcEditQuestion":
                                cId = chatQuestion.id;
                                if (cId === "tcQFeedback") {
                                    postChatObject.feedback = chatQuestion.nextSibling.childNodes[0].value; /*innerHTML;*/
                                } else {
                                    value = [];
                                    name = chatQuestion.innerHTML;
                                    value.push(chatQuestion.nextSibling.childNodes[0].value /*innerHTML*/ );
                                }
                                break;
                            default:
                                throw new Error("Chat Window Controller failed to render chat form data, invalid class name");
                                // break;
                        }

                        if (value !== "") {
                            postChatObject.questions.push({
                                question: name,
                                answers: value
                            });
                        }
                    }
                }

                chatWindowController.handlePostChatFilled(JSON.stringify(postChatObject));
            } else if (chatWindowController._chatState & State.kInExportForm) {

                let exportForm = inChatFormView.getInChatForm("tcExportForm");

                let email = exportForm.childNodes[0].childNodes[1].childNodes[0].value;

                // inSocket.send(TC.exportChatProtocol(mChatId, mCompanyId, email));
                // this._socket.send(TC.exportChatProtocol(mChatId, mCompanyId, email));
                // MGMG TODO -- implement this with WebSocketAsPromised

            } else if (chatWindowController._chatState === State.kInOfflineForm) {

                let offlineForm = inChatFormView.getInChatForm("tcFormEmailOfflineChat");

                let visitorEmail = offlineForm.childNodes[0].childNodes[1].childNodes[0].value;
                let visitorQuestion = offlineForm.childNodes[1].childNodes[1].childNodes[0].value;

                let frag = ETSChat.getStoredSessionData();

                let fragSettingsObj = JSON.parse(frag);

                let companyEmail = fragSettingsObj.chatInactiveButton.textEmailOffer;
                chatWindowController._companyId = fragSettingsObj.companyID;

                // MGMG TODO -- implement this with WebSocketAsPromised
                // this._socket.send(TC.exportOfflineQuestionProtocol(visitorEmail, companyEmail, visitorQuestion, kFragmentId, mCompanyId));

                document.getElementById("tcOfflineEmailAnswer").value = "";
                document.getElementById("tcOfflineQuestionAnswer").value = "";

            } else if (chatWindowController._chatState === State.kEnded) {
                if (message !== "") {

                    chatWindowController._chatFormView.setCloseButtonToEndButton();
                    chatWindowController._chatFormView.showUnpinButton();
                    let chat = JSON.parse(ETSChat.sessionStorage.read('chat'));
                    chat.ended = 0;
                    chat.chatStatus = 2;
                    this._chatState = State.kInChat;

                    chatWindowController._chatFormView.writeStatusMessage("You have re-opened the chat");

                    let frag = JSON.parse(TC.sessionStorage.read('tcFragSettings'));
                    this._companyid = frag.companyID;

                    ETSChat.sessionStorage.write('chat', JSON.stringify(chat));
                    // MGMG TODO -- implement this with WebSocketAsPromised
                    // inSocket.send(TC.createChatMessageProtocol(mChatId, mCompanyId, message, inFragmentId));
                    // createChatMessageMessage
                }
            }
        }

        addMessage(inMessage) {
            let messageObject = inMessage.package;

            let chat = JSON.parse(ETSChat.sessionStorage.read('chat'));

            // MGMG TODO -- fix the chat object
            // this._chat.messages.push(messageObject);
            // chat.messages.push(messageObject);
            // ETSChat.sessionStorage.write('chat', JSON.stringify(chat));
            // ETSChat.localStorage.write('chat', JSON.stringify(chat));

            // let localMessageCount = 0;
            // for (let j = 0; j < chat.messages.length; j++) {
            //     if (chat.messages[j].type !== 2) {
            //         localMessageCount++;
            //     }
            // }
            //
            // let serverMessagecount = messageObject.chatMessageCount;

/*
            if (localMessageCount != serverMessagecount) {
                console.log("local and server message count mismatch!");
                // MGMG TODO -- implement resync
                // inSocket.send(TC.createResyncRequestProtocol(mCompanyId, mChatId));
                return;
            }
*/


            /*
            var firstNameKey = messageObject.userID + "a";
            var lastNameKey  = messageObject.userID + "b";

            var fullNameKey  = messageObject.userID + "ab";

            if(agentDisplayFullNamesMap[fullNameKey] !== undefined &&
                agentDisplayFullNamesMap[fullNameKey] !== "") {
                agentDisplayName = agentDisplayFullNamesMap[fullNameKey];

            } else if(agentDisplayFirstNamesMap[firstNameKey] !== undefined &&
                agentDisplayFirstNamesMap[firstNameKey] !== "") {
                agentDisplayFirstName = agentDisplayFirstNamesMap[firstNameKey];

                agentDisplayLastName = agentDisplayLastNamesMap[lastNameKey];
                agentDisplayName = agentDisplayFirstName + " " + agentDisplayLastName;
            }
            */

            this._agentDisplayName = messageObject.senderName;

            let agentThumbnail = "";

            if (this._agentThumbnailsMap[messageObject.userID] !== undefined &&
                this._agentThumbnailsMap[messageObject.userID] !== "") {
                agentThumbnail = agentThumbnailsMap[messageObject.userID];
            }

            if (this._thumbnailsAllowed !== false) {
                if (agentThumbnail === "" && this._websiteDefaultThumbnail !== "")
                    agentThumbnail = this._websiteDefaultThumbnail;
            }

            let type = messageObject['type'];
            if (type & 1) //Normal messages are first bit up
                if (type == ChatMessageType.kVisitorMessage)
                    this._chatFormView.writeMessage(1, "You", messageObject['timestamp'], messageObject['message'])
                else {
                    this._chatFormView.writeMessage(0, agentDisplayName, messageObject['timestamp'], messageObject['message'], agentThumbnail);
                    this._chatFormView.setNotification("");
                }
            else { // status message are first bit down
                //if (type & 4) { // If this message is join message
                //	if (mChatJoined === true) { // Only ever show one join chat message for now
                //		break;
                //	}
                //
                //	mChatJoined = true;
                //}

                if (this._systemMessages)
                    this._chatFormView.writeStatusMessage(messageObject['message']);
            }

            if (this._minimised == 1) {
                this._chatFormView.minimiseChatWindow();
            }
        }


        renderChatFormData(inChatForm) {
            if (this._chatForm.childNodes.length) {
                this._chatFormView.insertBreakInChatForm();
                for (let i = 0; i < this._chatFormView.childNodes.length; i++) {
                    let chatQuestion = this._chatFormView.childNodes[i].childNodes[0];
                    let name = "";
                    let value = "";
                    switch (chatQuestion.className) {
                        case "tcRadioQuestion":
                            name = chatQuestion.innerHTML;
                            let inputNode = chatQuestion.nextSibling;
                            while (inputNode !== null) {
                                if (inputNode.childNodes[0].checked) {
                                    value = inputNode.childNodes[0].nextSibling.innerHTML;
                                    inputNode = null;
                                } else {
                                    inputNode = inputNode.nextSibling;
                                }
                            }
                            break;
                        case "tcEditQuestion":
                            name = chatQuestion.innerHTML;
                            value = chatQuestion.nextSibling.childNodes[0].innerHTML;
                            break;
                        default:
                            throw new Error("Chat Window Controller failed to render chat form data, invalid class name");
                            // break;
                    }
                    if (value !== "")
                        this._chatFormView.writeStatusMessage("<b>" + name + "</b><br />" + value);
                }
                this._chatFormView.insertBreakInChatForm();
            }
        }

        static createEndChatMessage(inChatId, inCompanyId) {
            var messageObject = {
                oty: 'tco',
                mty: 'message',
                rty: 'chat/visitorEnd',
                package: {
                    uuid: ETSChat.sessionStorage.read('uuid'),
                    sessionID: Session.getStoredSessionData().sessionID,
                    chatID: inChatId,
                    companyID: inCompanyId
                }
            }
            return messageObject;
        }

        static handleEndChatClicked() {
            // inSocket.send(TC.createEndChatProtocol(mChatId, mCompanyId));
            // MGMG TODO -- implement this with WebSocketAsPromised

            let endChatMessage = ChatWindowController.createEndChatMessage(ETSChat.chatWindowController._chatId, "");
            ETSChat.chatWindowController._wsp.open()
                .then(() => {
                    ETSChat.chatWindowController._wsp.sendRequest({message: endChatMessage})
                        .then( fulfill => {
                            console.log('fulfill request:', fulfill.message);
                            ETSChat.chatWindowController.handleChatEnded();
                        })
                        .catch((error) => {
                            console.log('catch error:', error);
                        });
                        // .finally(() => {
                        //     console.log('finally');
                        // });
                });

            ETSChat.chatWindowController.handleAgentStoppedTyping();
            let inChat = ETSChat.sessionStorage.write('inChat', 0);
            ETSChat.localStorage.write('chat', '');
            let cachedChat = ETSChat.sessionStorage.read('chat');
            if (cachedChat != null && cachedChat !== "") {
                cachedChat = JSON.parse(cachedChat);
                cachedChat.chatStatus = 3;
                ETSChat.sessionStorage.write('chat', JSON.stringify(cachedChat));
            }
        }

        static handleCloseChatClicked() {
            let chatWindowController = ETSChat.chatWindowController;
            if (chatWindowController._chatState === State.kInOfflineForm) {
                chatWindowController._chatFormView.clearChatWindow();
                chatWindowController._chatFormView.hideChatWindow();
                // inChatFormView.hideChatWindow();
                // MGMG TODO - implement TC.chatFinishedMessage() somehow
                // let showOfflineButton = TC.chatFinishedMessage();
                // MGMG TODO -- showOfflineButton
                // TC.EventMediator.mediateEvent(showOfflineButton);
                chatWindowController._chatState = State.kNone;
                chatWindowController.onChatFinished();
            } else if (chatWindowController._chatState === State.kEnded) {
                // MGMG TODO -- implement this with WebSocketAsPromised
                // inSocket.send(TC.createChatClosedProtocol());
                // let inChat = ETSChat.sessionStorage.write('inChat', 0);
                chatWindowController._chatButtonNotClicked = true;
                chatWindowController.onChatFinished();
            } else if (this._chatState === State.kInPostChat) {
                // MGMG TODO -- implement this with WebSocketAsPromised
                // inSocket.send(TC.createChatClosedProtocol());
                let inChat = ETSChat.sessionStorage.write('inChat', 0);
                chatWindowController._chatButtonNotClicked = true;
                chatWindowController.onChatFinished();
            } else if (this._chatState === State.kInExportForm) {
                // MGMG TODO -- implement this with WebSocketAsPromised
                // inSocket.send(TC.createChatClosedProtocol());
                let inChat = ETSChat.sessionStorage.write('inChat', 0);
                chatWindowController._chatButtonNotClicked = true;
                chatWindowController.onChatFinished();
            }
        }

        handleUnpinClicked() {
            // MGMG TODO -- implement this for new server
            window.open("https://www.etschat.com/script/TonicChatNewWindow.html?wfid=" + inFragmentId + "&chid=" + mChatId +
                "&sid=" + mSessionId + "&vid=" + JSON.parse(TC.cookieStorage.read(inFragmentId)).visitorID, "ETSChat, powered by Engage To Sell LLC", "height=600, width=800");
            this._chatFormView.minimiseChatWindow();
            //mChatState = State.kNone;
        }

        handlePrintClicked() {
            let children = mChat.messages;
            let txtTranscript = document.createElement('div');


            for (let i = 0; i < children.length; i++) {
                let date = new Date(children[i].timestamp);
                let hours = date.getHours();
                let minutes = date.getMinutes();
                let seconds = date.getSeconds();
                let formattedTime = hours + ':' + minutes + ':' + seconds;

                if (children[i].type == 5)
                    txtTranscript.innerHTML += "<p>" + "You, " + formattedTime + "<br>" + children[i].message + "</p>";
                else
                    txtTranscript.innerHTML += "<p>" + "Agent, " + formattedTime + "<br>" + children[i].message + "</p>";
            }

            let newwindow = window.open();
            newwindow.document.body.appendChild(txtTranscript);
            newwindow.print();
            newwindow.close();
        }

        handleExportClicked() {
            if (this._chat.messages.length) {
                let chatForm = ETSChat.createChatFormObject();
                this._chatFormView.insertBreakInChatForm();
                chatForm.chatFormIntroOne = "Export Chat Form";
                let questionForForm = TC.createChatFormQuestion();
                questionForForm.question = "Please enter your email below."
                questionForForm.optionType = 0x01;

                chatForm.chatQuestions.push(questionForForm);
                this._chatFormView.hideMessageArea();
                this._chatFormView.hidePrintButton();
                this._chatFormView.hideExportButton();
                this._previousChatState = this._chatState;
                this._chatState = State.kInExportForm;
                this._chatFormView.createInChatForm("tcExportForm", chatForm);

            } else {
                this._chatFormView.writeStatusMessage("Unable to export chat, as there are no chat messages.");
            }
        }

        handleVisitorTypingKeyPress(message) {
            if (this._chatState & 1) {
                // MGMG TODO -- implement this with WebSocketAsPromised
                // inSocket.send(TC.createVisitorTypingProtocol(this._companyId, this._chatId, message));
            }
        }

        handleMinimiseClicked() {
            if (this._chatState === State.kEnded) {
                this._minimised = 0;
                this._chatFormView.expandChatWindow();
                handleCloseChatClicked();
            } else {
                thils._minimised = 1;
            }
        }

        handleExpandedClicked() {
            this._minimised = 0;
        }

        createNewChatRequest(inInitialMessage) {
            let fragSettings = Session.getStoredSessionData();
            let messageObject = {
                oty: 'tco',
                mty: 'message',
                rty: 'chat/new',
                package: {
                    sessionID: fragSettings.sessionID,
                    visitorID: fragSettings.visitorID,
                    uuid: ETSChat.sessionStorage.read('uuid'),
                    fragmentID: this._fragmentId,
                    chatID: this._chatId,
                    initialMessage: inInitialMessage,
                    smsPhoneNumber: "",
                    smsFirstName: "",
                    smsLastName: "",
                    smsMessage: "",
                    isBotSession: false
                }
            };
            return messageObject;
        }

        //PRIVATE METHODS
        handleChatStarted(inInitialMessage) {
            this._chatFormView.enableChat();
            this._chatFormView.showMessageArea();
            this._chatFormView.createIntroHeader();
            this._chatFormView.showExportButton();
            this._chatFormView.showUnpinButton();
            this._chatFormView.showPrintButton();
            this._chatFormView.setCloseButtonToEndButton();
            ETSChat.localStorage.write('chat', '');
            let createNewChatMessage = this.createNewChatRequest(null);
            if(createNewChatMessage) {
                this._wsp.open()
                    .then(() => {
                        this._wsp.sendRequest({message: createNewChatMessage})
                            .then( fulfill => {
                                this._chatId = fulfill.message.package.chatID;
                                if(fulfill.message.package.chat !== null && fulfill.message.package.chat !== undefined) {
                                    this._chat = JSON.parse(fulfill.message.package.chat);
                                }
                                this._chat.chatStatus = 1;
                                this._chatState = State.kInChat;

                                ETSChat.sessionStorage.write("chat", JSON.stringify(this._chat));
                                ETSChat.localStorage.write("chat", JSON.stringify(this._chat));
                                console.log('fulfill request:', fulfill.message);
                                // this.handleChatStarted();
                            })
                            .catch((error) => {
                                console.log('catch error:', error);
                            });
                            // .finally(() => {
                            //     console.log('finally');
                            // });
                    });

            }
            //TODO initial message.
            // MGMG TODO implement this with WebSocketAsPromised

            // inSocket.send(TC.createNewChatRequestProtocol(inFragmentId, inInitialMessage, mChatId, TC.EtsBotSession));
            // send over possible existing chat Id from prechat
        }

        handlePreChatFilled(inPreChatData) {
            // MGMG TODO implement with WebSocketAsPromised
            // inSocket.send(TC.createPreChatFilledProtocol(inFragmentId, mChatId, inPreChatData));
        }

        handlePostChatFilled(inPostChatData) {
            // MGMG TODO implement with WebSocketAsPromised
            // inSocket.send(TC.createPostChatFilledProtocol(inFragmentId, mChatId, inPostChatData));
        }

        handleChatPreChat() {
            // MGMG TODO implement with WebSocketAsPromised
            // inSocket.send(TC.createPreChatRequestProtocol(inFragmentId));
        }

        handleChatResurrected(inChat, inChatUserIds, inChatThumbnails, defaultThumbnail,
                                       agentNameIds, agentFirstNames, agentLastNames, displayFullNames) {

            let isStandaloneChatWindow = TC.checkVisitorIsOnMobile() === true && kIsStandaloneWindow === false;
            let frag = JSON.parse(TC.sessionStorage.read('tcFragSettings'));
            let displayEmbeddedChatWindowOnMobile = frag.chatWindow.displayEmbeddedChatWindowOnMobile;
            if (TC.checkVisitorIsOnMobile() === true && displayEmbeddedChatWindowOnMobile === true) {
                isStandaloneChatWindow = false;
            }


            if (this._chat === null || this._chat === undefined) {
                throw new Error("Failed to resurrect chat as the chat object is null or undefined.");
            } else if (isStandaloneChatWindow === false) {
                let chat = JSON.parse(ETS.sessionStorage.read('chat'));
                ETS.sessionStorage.write('chat', JSON.stringify(this._chat));
                if (!this._chat.ended) {
                    this._chatState = State.kInChat;

                    if (this._chatState.chatStartType === 3)
                        ETSChat.sessionStorage.write('autoInviteOn', 1);

                    ETSChat.sessionStorage.write('inChat', 1);

                    let fragSettings = Session.getStoredSessionData();
                    let agentThumbnailEnabled = fragSettings.chatWindow.displayAgentThumb;

                    let thumbnailRez = 0;

                    //sets the users thumbnail to a map from server if user has exited browser only.
                    if (agentThumbnailEnabled === true && this._chatUserIds !== undefined) {
                        for (let i = 0; i < this._chatUserIds.length; i++) {
                            if (defaultThumbnail !== "") {
                                inChatThumbnails[i] = defaultThumbnail;
                            }

                            this._agentThumbnailsMap[inChatUserIds[i]] = inChatThumbnails[i];
                            ETSChat.sessionStorage.write(inChatUserIds[i], inChatThumbnails[i]);
                            thumbnailRez = 1;
                        }
                    }

                    let displayAgentNameEnabled = fragSettings.chatWindow.displayAgentName;
                    let displayNameRez = 0;

                    //For complete res from the server only hits if browser has been exited completely.
                    if (agentNameIds !== undefined) {
                        if (displayAgentNameEnabled === true && agentNameIds.length > 0) {
                            for (let i = 0; i < agentNameIds.length; i++) {
                                if (displayFullNames[i] !== "") {
                                    this._agentDisplayFullNamesMap[agentNameIds[i] + "ab"] = displayFullNames[i];

                                    ETSChat.sessionStorage.write(agentNameIds[i] + "ab", displayFullNames[i]);
                                } else {
                                    agentFirstNames[i] = agentFirstNames[i].charAt(0).toUpperCase() + agentFirstNames[i].slice(1);
                                    agentDisplayFirstNamesMap[agentNameIds[i] + "a"] = agentFirstNames[i];
                                    agentLastNames[i] = agentLastNames[i].charAt(0).toUpperCase() + agentLastNames[i].slice(1);
                                    agentDisplayLastNamesMap[agentNameIds[i] + "b"] = agentLastNames[i];

                                    ETSChat.sessionStorage.write(agentNameIds[i] + "a", agentFirstNames[i]);
                                    ETSChat.sessionStorage.write(agentNameIds[i] + "b", agentLastNames[i]);
                                }
                                displayNameRez = 1;
                            }
                        }
                    }


                    if (displayAgentNameEnabled === false) {
                        this._agentDisplayName = "Agent";
                    }

                    this._chatFormView.enableChat();
                    this._chatFormView.showChatWindow();
                    this._chatFormView.createIntroHeader();

                    let agentThumbnail = "";

                    let mapEntryFound = 0;


                    for (let i = 0; i < this._chat.messages.length; i++) {
                        let chatMessage = this._chat.messages[i];

                        if (displayAgentNameEnabled === true && displayNameRez === 0 && chatMessage.type === 1) {
                            let displayNameKey = chatMessage.userID + "ab";
                            let displayFullName = ETSChat.sessionStorage.read(displayNameKey);
                            if (displayFullName !== "" && displayFullName !== undefined) {
                                if (this._agentDisplayFullNamesMap[displayNameKey] === undefined ||
                                    this._agentDisplayFullNamesMap[displayNameKey] === "") {

                                    this._agentFullDisplayName = TC.sessionStorage.read(displayNameKey);
                                    this._agentDisplayFullNamesMap[displayNameKey] = agentFullDisplayName;
                                    this._agentDisplayName = agentFullDisplayName;
                                } else {
                                    this._agentDisplayName = this._agentDisplayFullNamesMap[displayNameKey];
                                }
                            } else {
                                let firstNameKey = chatMessage.userID + "a";
                                let lastNameKey = chatMessage.userID + "b";

                                if (this._agentDisplayFirstNamesMap[firstNameKey] === undefined ||
                                    this._agentDisplayFirstNamesMap === "") {
                                    this._agentDisplayFirstName = ETSChat.sessionStorage.read(firstNameKey);
                                    this._agentDisplayLastName = ETSChat.sessionStorage.read(lastNameKey);

                                    this._agentDisplayFirstNamesMap[firstNameKey] = this._agentDisplayFirstName;
                                    this._agentDisplayLastNamesMap[lastNameKey] = this._agentDisplayLastName;
                                    this._agentDisplayName = this._agentDisplayFirstName + " " + this._agentDisplayLastName;

                                } else {
                                    this._agentDisplayName = this._agentDisplayFirstNamesMap[firstNameKey] + " " +
                                        this._agentDisplayLastNamesMap[lastNameKey];
                                }
                            }
                        } else {

                            let displayNameKey = chatMessage.userID + "ab";
                            if (this._agentDisplayFullNamesMap[displayNameKey] !== undefined && displayAgentNameEnabled === true) {
                                this._agentDisplayName = this._agentDisplayFullNamesMap[displayNameKey];

                            } else if (displayAgentNameEnabled === true && chatMessage.type === 1) {

                                let firstNameKey = chatMessage.userID + "a";
                                let lastNameKey = chatMessage.userID + "b";

                                this._agentDisplayName = agentDisplayFirstNamesMap[firstNameKey] + " " +
                                    this._agentDisplayLastNamesMap[lastNameKey];
                            }
                        }

                        this._agentDisplayName = chatMessage.senderName;

                        //if the user has not exited the browser get the thumbnail from session store
                        if (agentThumbnailEnabled === true && thumbnailRez === 0) {
                            if (this._agentThumbnailsMap[chatMessage.userID] === undefined ||
                                this._agentThumbnailsMap[chatMessage.userID] === "") {
                                agentThumbnail = TC.sessionStorage.read(chatMessage.userID);
                                this._agentThumbnailsMap[chatMessage.userID] = agentThumbnail;
                            } else {
                                agentThumbnail = this._agentThumbnailsMap[chatMessage.userID];
                            }
                        } else {
                            agentThumbnail = this._agentThumbnailsMap[chatMessage.userID];
                        }

                        var type = chatMessage.type;
                        if (type & 1) { //normal message.
                            (type === ETSChat.ChatMessageType.kVisitorMessage) ?
                                this._chatFormView.writeMessage(1, "You", chatMessage.timestamp, chatMessage.message): inChatFormView.writeMessage(0, agentDisplayName, chatMessage.timestamp, chatMessage.message, agentThumbnail);
                        } else { //sys
                            if (this._systemMessages)
                                this._chatFormView.writeStatusMessage(chatMessage.message);
                        }
                    }
                    this._chatId = inChat._id;
                    this._chat = inChat;
                    this._companyId = inChat.companyID;
                    //chat in event message therefore let the view know to build the chat from the chat data.
                } else {
                    // MGMG TODO -- implement outside of event loop
                    // TC.EventMediator.mediateEvent(TC.showChatButtonMessage());
                }
            }
        }

        handleShowChatWindow() {
            this._chatFormView.clearChatWindow();
            this._chatFormView.enableChat();
            this._chatFormView.showMessageArea();
            this._chatFormView.createIntroHeader();
            this._chatFormView.showExportButton();
            this._chatFormView.showUnpinButton();
            this._chatFormView.showPrintButton();
            this._chatFormView.setCloseButtonToEndButton();
            // MGMG TODO implement outside of event loop
            // TC.EventMediator.mediateEvent(TC.chatStartedMessage());

            this.showChatWindow();
        }

        handleChatButtonClicked() {
            this._chatState = State.kNone;
            if (this._chatState === State.kNone) {
                if (this._modulesEnabled >> 0x02 & 1) {
                    if (this._preChatFormSettings === null) {
                        throw new Error("Failed to create pre chat form, as the pre chat form settings object is null");
                    } else {
                        this.handleChatPreChat();
                    }
                } else {
                    this.handleChatStarted(null);
                }

                this._chatFormView.clearChatWindow();

                this._chatButtonNotClicked = false;

                this._chatFormView.enableChat();
                this._chatFormView.showMessageArea();
                this._chatFormView.createIntroHeader();
                this._chatFormView.showExportButton();
                this._chatFormView.showUnpinButton();
                this._chatFormView.showPrintButton();
                this._chatFormView.setCloseButtonToEndButton();
                ETSChat.localStorage.write('chat', '');

                // MGMG TODO implement outside of event loop
                // TC.EventMediator.mediateEvent(TC.chatStartedMessage());
                this.sendChatStartedMessage();
                this.showChatWindow();
            }

        }

        sendChatStartedMessage() {

        }

        handleVisitorInviteClicked() {
            //if invited it should go straight to chat.
            /*if (mModulesEnabled >> 0x02 & 1) {
                if (mPreChatFormSettings === null) {
                    throw new Error("Failed to create pre chat form, as the pre chat form settings object is null");
                } else {
                    handleChatPreChat();
                }
            }*/

            this._chat = mChatInvite;
            this._chatId = mChatInviteId;

            //inChatFormView.clearChatWindow();
            this._chatFormView.enableChat();
            this._chatFormView.showMessageArea();
            this._chatFormView.createIntroHeader();
            this._chatFormView.showExportButton();
            this._chatFormView.showUnpinButton();
            this._chatFormView.showPrintButton();
            this._chatFormView.setCloseButtonToEndButton();

            this._chatState = State.kInChat;

            this._chat.chatStatus = 2;

            ETSChat.sessionStorage.write('chat', JSON.stringify(this._chat));

            // MGMG TODO implement outside of event loop
            // TC.EventMediator.mediateEvent(TC.chatStartedMessage());

            this.showChatWindow();
        }

        showChatWindow() {
            this._chatFormView.enableSendButton();
            this._chatFormView.showSendButton();
            this._chatFormView.showUnpinButton();
            this._chatFormView.showChatWindow();
        }

        handleChatEnabled(fragmentSettings) {
            this._modulesEnabled = fragmentSettings.modulesEnabled;
            this._sessionId = fragmentSettings.sessionID;
            this._chatFormView.setUISettings(fragmentSettings.chatWindow);
            this._systemMessages = fragmentSettings.chatWindow.displayStatusMessages;

            if (this._modulesEnabled >> 0x02 & 1) //PreChatEnabled
                this._preChatFormSettings = fragmentSettings.preChatForm;
            if (this._modulesEnabled >> 0x03 & 1) //PostChatEnabled
                this._postChatFormSettings = fragmentSettings.postChatForm;
        }

        handleChatEnded() {
            this._chatFormView.writeStatusMessage("The chat has ended.");
            //inChatFormView.disableChat();
            this._chatFormView.changeEndButtonToClose();
            this._chatFormView.setEndButtonToCloseButton();
            //inChatFormView.hideMessageArea();
            this._chatFormView.hideUnpinButton();
            this._chatState = State.kEnded;

            if (this._modulesEnabled >> 0x03 & 1) {
                if (this._postChatFormSettings === null) {
                    throw new Error("Failed to create post chat form, as the post chat form settings object is null");
                } else {
                    // MGMG TODO -- implement with WebSocketAsPromised
                    // inSocket.send(TC.createPostChatRequestProtocol(inFragmentId, this._chatId));
                    // this.handleCloseChatClicked()
                }
            } else {
                // this.handleCloseChatClicked()
                if (this._minimised) {
                    //handleCloseChatClicked();
                } else {
                    //inChatFormView.hideSendButton();
                }
            }
        }

        //PUBLIC METHODS
        handleRequest(inRequestType, inMessageObject) {
            let chatForm;
            switch (inRequestType) {
                case "chat/join":
                    let fragSettings = ETSChat.getStoredSessionData();
                    this._websiteDefaultThumbnail = fragSettings.websiteDefaultThumbnail;
                    this._thumbnailsAllowed = fragSettings.chatWindow.displayAgentThumb;

                    let displayName = inMessageObject.displayName;

                    if (fragSettings.chatWindow.displayAgentName == true && displayName == "") {
                        this._agentDisplayFirstName = inMessageObject.firstName;
                        this._agentDisplayLastName = inMessageObject.lastName;
                        this._agentDisplayName = inMessageObject.firstName + " " + inMessageObject.lastName;
                    } else if (fragSettings.chatWindow.displayAgentName === true && displayName !== "") {
                        this._agentDisplayName = displayName;
                    } else {
                        this._agentDisplayName = "Agent";
                    }

                    if (fragSettings.chatWindow.displayAgentName === true)
                        this.handleSetAgentDisplayName(displayName, this._agentDisplayFirstName, this._agentDisplayLastName,
                            inMessageObject.userID);


                    if (this._websiteDefaultThumbnail !== "")
                        inMessageObject.profPicture = this._websiteDefaultThumbnail;
                    if (this._thumbnailsAllowed === true)
                        this.handleSetAgentThumbnail(inMessageObject);

                    break;
                case "chat/agentTyping":
                    this.handleAgentTyping();
                    break;
                case "chat/agentStoppedTyping":
                    this.handleAgentStoppedTyping();
                    break;

                case "chat/resync":
                    //console.log("chat resync received : whole chat sent down is::::");
                    //console.log(inMessageObject.chat);
                    let resyncedChat = JSON.parse(inMessageObject.chat);
                    //console.log(chat);

                    this._chat.messages = [];
                    ETSChat.sessionStorage.write('chat', JSON.stringify(resyncedChat));
                    ETSChat.localStorage.write('chat', JSON.stringify(resyncedChat));

                    this._chatFormView.clearChatWindow();
                    this._chatFormView.createIntroHeader();

                    for (let i = 0; i < resyncedChat.messages.length; i++) {
                        let chatMessage = resyncedChat.messages[i];
                        this._chat.messages.push(chatMessage);

                        this._agentDisplayName = chatMessage.senderName;

                        let agentThumbnail = "";

                        if (this._agentThumbnailsMap[chatMessage.userID] !== undefined &&
                            this._agentThumbnailsMap[chatMessage.userID] !== "") {
                            this._agentThumbnail = this._agentThumbnailsMap[chatMessage.userID];
                        }

                        if (this._thumbnailsAllowed !== false) {
                            if (agentThumbnail === "" && this._websiteDefaultThumbnail !== "")
                                agentThumbnail = this._websiteDefaultThumbnail;
                        }

                        let type = chatMessage.type;
                        if (type === ChatMessageType.kVisitorMessage)
                            this._chatFormView.writeMessage(1, "You", chatMessage.timestamp, chatMessage.message)
                        else if (type === ChatMessageType.kText || type === ChatMessageType.kLiveInviteMessage) {
                            this._chatFormView.writeMessage(0, this._agentDisplayName, chatMessage.timestamp, chatMessage.message, agentThumbnail);
                            this._chatFormView.setNotification("");
                        } //else { // status message are first bit down
                        //if (type & 4) { // If this message is join message
                        //	if (mChatJoined === true) { // Only ever show one join chat message for now
                        //		break;
                        //	}
                        //
                        //	mChatJoined = true;
                        //}

                        /*
                        kText                   : 0x01,
                        kLiveInviteMessage      : 0x03,
                        kJoinedChat             : 0x04,
                        kVisitorMessage         : 0x05,
                        kLeftChat               : 0x08,
                        */
                        // if (mSystemMessages)
                        //     inChatFormView.writeStatusMessage(chatMessage.message);
                        //}

                        if (this._minimised == 1) {
                            this._chatFormView.minimiseChatWindow();
                        }
                    }

                    break;

                case "chat/message":
                    let messageObject = JSON.parse(inMessageObject.message);

                    let chat = JSON.parse(ETSChat.sessionStorage.read('chat'));

                    this._chat.messages.push(messageObject);
                    chat.messages.push(messageObject);
                    ETSChat.sessionStorage.write('chat', JSON.stringify(chat));
                    ETSChat.localStorage.write('chat', JSON.stringify(chat));

                    let localMessageCount = 0;
                    for (let j = 0; j < chat.messages.length; j++) {
                        if (chat.messages[j].type !== 2) {
                            localMessageCount++;
                        }
                    }

                    let serverMessagecount = messageObject.chatMessageCount;

                    if (localMessageCount !== serverMessagecount) {
                        console.log("local and server message count mismatch!");
                        // MGMG TODO implement
                        // inSocket.send(TC.createResyncRequestProtocol(mCompanyId, mChatId));
                        return;
                    }


                    /*
                var firstNameKey = messageObject.userID + "a";
                var lastNameKey  = messageObject.userID + "b";

                var fullNameKey  = messageObject.userID + "ab";

                if(agentDisplayFullNamesMap[fullNameKey] !== undefined &&
                    agentDisplayFullNamesMap[fullNameKey] !== "") {
                    agentDisplayName = agentDisplayFullNamesMap[fullNameKey];

                } else if(agentDisplayFirstNamesMap[firstNameKey] !== undefined &&
                    agentDisplayFirstNamesMap[firstNameKey] !== "") {
                    agentDisplayFirstName = agentDisplayFirstNamesMap[firstNameKey];

                    agentDisplayLastName = agentDisplayLastNamesMap[lastNameKey];
                    agentDisplayName = agentDisplayFirstName + " " + agentDisplayLastName;
                }
				*/

                    this._agentDisplayName = messageObject.senderName;

                    let agentThumbnail = "";

                    if (this._agentThumbnailsMap[messageObject.userID] !== undefined &&
                        this._agentThumbnailsMap[messageObject.userID] !== "") {
                        agentThumbnail = this._agentThumbnailsMap[messageObject.userID];
                    }

                    if (this._thumbnailsAllowed !== false) {
                        if (agentThumbnail === "" && this._websiteDefaultThumbnail !== "")
                            agentThumbnail = this._websiteDefaultThumbnail;
                    }

                    let type = messageObject['type'];
                    if (type & 1) //Normal messages are first bit up
                        if (type === ChatMessageType.kVisitorMessage)
                            this._chatFormView.writeMessage(1, "You", messageObject['timestamp'], messageObject['message'])
                        else {
                            this._chatFormView.writeMessage(0, this._agentDisplayName, messageObject['timestamp'], messageObject['message'], agentThumbnail);
                            this._chatFormView.setNotification("");
                        }
                    else { // status message are first bit down
                        //if (type & 4) { // If this message is join message
                        //	if (mChatJoined === true) { // Only ever show one join chat message for now
                        //		break;
                        //	}
                        //
                        //	mChatJoined = true;
                        //}

                        if (this._systemMessages)
                            this._chatFormView.writeStatusMessage(messageObject['message']);
                    }

                    if (this._minimised === 1) {
                        this._chatFormView.minimiseChatWindow();
                    }
                    break;
                case "chat/end":
                    chat = JSON.parse(TC.sessionStorage.read('chat'));
                    if (chat != null) {
                        chat.ended = 1;
                        ETSChat.sessionStorage.write('chat', JSON.stringify(chat));
                        this.handleChatEnded();
                    }
                    break;
                case "chat/reopened":
                    //The agent Reopened the chat!
                    if (this._chatState === State.kEnded || this._chatState === State.kNone ||
                        this._chatState === State.kInPostChat) {
                        this._chatFormView.showMessageArea();
                        this._chatFormView.showExportButton();
                        this._chatFormView.showUnpinButton();
                        this._chatFormView.showPrintButton();
                        this._chatFormView.showSendButton();
                        this._chatFormView.clearChatWindow();

                        this._chatFormView.setCloseButtonToEndButton();
                        this._chatFormView.removeInChatForm("tcPostChatForm");

                        let chat = JSON.parse(ETSChat.sessionStorage.read('chat'));
                        chat.ended = 0;
                        chat.chatStatus = 2;
                        this.handleChatResurrected(chat);

                        ETSChat.sessionStorage.write('chat', JSON.stringify(chat));

                        this._chatFormView.writeStatusMessage("The chat has been re-opened.");
                        this._chatState = State.kInChat;
                    }
                    break;
                case "chat/new":

                    if (this._chatState == State.kInChat)
                        break;

                    this._chatJoined = false;
                    this._chatEnded = 0;
                    this._chat = JSON.parse(inMessageObject.chat);

                    // if mChat has smsVisitorNumber it means it's SMS form response.
                    if(this._chat.smsVisitorNumber)
                        break;

                    this._chatId = this._chat._id;
                    this._companyId = this._chat.companyID;
                    this._chat.ended = 0;

                    if (this._chat.chatStatus !== 7) {
                        if (this._chatButtonNotClicked === true) {
                            // MGMG TODO -- impement outside of event loop
                            // TC.EventMediator.mediateEvent(TC.hideChatButtonMessage());

                            if (!ETSUtils.checkVisitorIsOnMobile()) {
                                this.showChatWindow();
                            }
                        }

                        this._chat.chatStatus = 1;
                        this._chatState = State.kInChat;
                    } else {
                        this._chatState = State.kVisitorInvitePending;
                    }

                    ETSChat.sessionStorage.write('chat', JSON.stringify(mChat));


                    if (this._chat.chatStatus === 7) {
                        this._chatInvite = mChat;
                        this._chatInviteId = mChatId;
                        // MGMG TODO -- implement without event loop
                        // TC.EventMediator.mediateEvent(TC.inviteVisitorToChatMessage());
                    } else {
                        if (this._chatState === State.kNone)
                            showChatWindow();

                        this._chatFormView.clearChatWindow();

                        this._chatFormView.enableChat();
                        this._chatFormView.showMessageArea();
                        this._chatFormView.createIntroHeader();
                        this._chatFormView.showExportButton();
                        this._chatFormView.showUnpinButton();
                        this._chatFormView.showPrintButton();
                        this._chatFormView.setCloseButtonToEndButton();

                        if (this._chat.messages.length) {
                            let firstMsg = this._chat.messages[0];
                            this._chatFormView.writeMessage(1, "Visitor", firstMsg.timestamp, firstMsg.message);
                        }

                        // MGMG TODO implement without event loop
                        // TC.EventMediator.mediateEvent(TC.chatStartedMessage());

                        if (kIsStandaloneWindow === true) {
                            window.location.search += '&chid=' + this._chatId;
                        }
                    }

                    break;
                case "chat/prechat":
                    if (this._chatState === State.kNone)
                        this.showChatWindow();

                    this._chatFormView.clearChatWindow();

                    this._chatState = State.kInPreChat;
                    chatForm = TC.createChatFormObject();

                    chatForm.chatFormIntroOne = "Start a Live Chat";
                    chatForm.chatFormIntroTwo = "Please complete the form below to get started";

                    if (this._preChatFormSettings.askForEmail) {
                        let askForEmailQuestion = ETSUtils.createChatFormQuestion();
                        askForEmailQuestion.id = "tcQEmail";
                        askForEmailQuestion.optionType = 0x01;
                        askForEmailQuestion.question = "Please enter your email address.";
                        chatForm.chatQuestions.push(askForEmailQuestion);
                    }

                    if (this._preChatFormSettings.askForName) {
                        let askForNameQuestion = ETSUtils.createChatFormQuestion();
                        askForNameQuestion.id = "tcQNameFirst";
                        askForNameQuestion.optionType = 0x01;
                        askForNameQuestion.question = "Please enter your first name.";
                        chatForm.chatQuestions.push(askForNameQuestion);

                        askForNameQuestion = ETSUtils.createChatFormQuestion();
                        askForNameQuestion.id = "tcQNameLast";
                        askForNameQuestion.optionType = 0x01;
                        askForNameQuestion.question = "Please enter your last name.";
                        chatForm.chatQuestions.push(askForNameQuestion);
                    }

                    let newVisitor = ETSChat.sessionStorage.read("newVisitor");

                    for (let i = 0; i < this._preChatFormSettings.preChatQuestions.length; i++) {

                        let preChatQuestion = mPreChatFormSettings.preChatQuestions[i];

                        let onReturningUser = preChatQuestion.onReturningUser;
                        let onNewUser = preChatQuestion.onNewUser;

                        if ((!onNewUser && !onReturningUser) || (newVisitor === "true" && onNewUser) ||
                            (newVisitor === "false" && onReturningUser)) {
                            let questionForForm = TC.createChatFormQuestion();

                            questionForForm.optionType = preChatQuestion.optionType;
                            questionForForm.question = preChatQuestion.question;

                            for (let j = 0; j < preChatQuestion.options.length; j++)
                                questionForForm.options.push(preChatQuestion.options[j]);


                            chatForm.chatQuestions.push(preChatQuestion);
                        }
                    }

                    if (this._preChatFormSettings.askForInitialQuestion) {
                        let askForInitialQuestion = ETSChat.createChatFormQuestion();
                        askForInitialQuestion.id = "tcQInitialQuestion";
                        askForInitialQuestion.optionType = 0x01;
                        askForInitialQuestion.question = mPreChatFormSettings.textAskForInitialQuestion;
                        chatForm.chatQuestions.push(askForInitialQuestion);
                    }

                    this._chatFormView.createInChatForm("tcPreChatForm", chatForm);
                    this._chatFormView.hideMessageArea();
                    this._chatFormView.hidePrintButton();
                    this._chatFormView.hideExportButton();
                    this._chatFormView.setEndButtonToCloseButton();
                    this._chatId = inMessageObject.chatID;

                    /*
                     * TEdwards commented out as this breaks a lot of things.
                     * Has been changed in business logic with the above code.
                     *mChatEnded  = 0;
                    mChat       = JSON.parse(inMessageObject.chat);

                    // assign chatId so when state goes to new chat, server will know chatId already assigned
                    mChatId     = mChat._id;
                    mCompanyId  = mChat.companyID;

                    TC.sessionStorage.write('chat', JSON.stringify(mChat));*/

                    break;
                case "chat/postchat":
                    this._chatFormView.enableSendButton();
                    if (this._minimised) {
                        this._chatFormView.expandChatWindow();
                    }

                    this._chatState = State.kInPostChat;

                    chatForm = ETSChat.createChatFormObject();

                    chatForm.chatFormIntroOne = "Thank you for the chat!";
                    chatForm.chatFormIntroTwo = "Feel free to leave us any additional feedback.";

                    if (this._postChatFormSettings.postChatRating) {
                        let questionForForm = ETSUtils.createChatFormQuestion();
                        questionForForm.id = "tcQChatRating";
                        questionForForm.question = "How would you rate the support you have received?";
                        questionForForm.optionType = 0x02;
                        questionForForm.options.push("Excellent");
                        questionForForm.options.push("Good");
                        questionForForm.options.push("Average");
                        questionForForm.options.push("Poor");
                        chatForm.chatQuestions.push(questionForForm);
                    }

                    newVisitor = TC.sessionStorage.read("newVisitor");

                    for (let i = 0; i < this._postChatFormSettings.postChatQuestions.length; i++) {

                        let postChatQuestion = this._postChatFormSettings.postChatQuestions[i];

                        let onReturningUser = postChatQuestion.onReturningUser;
                        let onNewUser = postChatQuestion.onNewUser;

                        if ((!onNewUser && !onReturningUser) || (newVisitor === "true" && onNewUser) ||
                            (newVisitor === "false" && onReturningUser)) {
                            let questionForForm = ETSUtils.createChatFormQuestion();

                            questionForForm.optionType = postChatQuestion.optionType;
                            questionForForm.question = postChatQuestion.question;

                            for (let j = 0; j < postChatQuestion.options.length; j++)
                                questionForForm.options.push(postChatQuestion.options[j]);

                            chatForm.chatQuestions.push(postChatQuestion);
                        }
                    }

                    if (this._postChatFormSettings.postChatAskFeedback) {
                        let questionForForm = TC.createChatFormQuestion();
                        questionForForm.id = "tcQFeedback";
                        questionForForm.optionType = 0x01;
                        questionForForm.question = "Feedback";
                        chatForm.chatQuestions.push(questionForForm);
                    }

                    this._chatFormView.clearChatWindow();

                    this._chatFormView.createInChatForm("tcPostChatForm", chatForm);

                    if (this._postChatFormSettings.postChatAutoClose) {
                        this._postChatTimer.setPeriod(this._postChatFormSettings.postChatAutoCloseTime);
                        this._postChatTimer.setEnabled(1);
                    }

                    this._chatState = State.kInPostChat;
                    this._chatFormView.setEndButtonToCloseButton();

                    break;
                case "chat/postChatFilled":
                    this.onChatFinished();
                    break;
                case "chat/export":
                    this._chatState = this._previousChatState;

                    if (inMessageObject.error === undefined)
                        this._chatFormView.writeStatusMessage("An email has been sent to your inbox with the chat transcript.");
                    else
                        this._chatFormView.writeStatusMessage(inMessageObject.error);


                    this._chatFormView.removeInChatForm("tcExportForm");

                    this._chatFormView.showPrintButton();
                    this._chatFormView.showExportButton();

                    if (mChatState === State.kInChat)
                        this._chatFormView.showMessageArea();
                    break;
                case "chat/offlineQuestion":
                    let divTagCheck = document.getElementById("offlineFormErrorMessage");
                    if (inMessageObject.error === undefined) {
                        if (divTagCheck != null)
                            document.getElementById("offlineFormErrorMessage").remove();
                        this._chatFormView.writeStatusMessage("An email has been sent to us with your query, thank you.", "offlineFormErrorMessage");
                        document.getElementById("tcBusinessHoursQuestionLabel").remove();
                        document.getElementById("tcFormEmailOfflineChat").remove();
                        this._chatFormView.hideSendButton();
                    } else {
                        if (divTagCheck != null)
                            document.getElementById("offlineFormErrorMessage").remove();
                        this._chatFormView.writeStatusMessage(inMessageObject.error, "offlineFormErrorMessage");
                    }
                    break;
                case "visInvite/check":
                    if (inMessageObject.invited === true && this._chatState !== State.kInChat) {
                        this._chatInvite = JSON.parse(inMessageObject.chat);
                        this._chatInviteId = mChatInvite._id;
                        this._companyId = mChatInvite.companyID;
                        //mChatState  = 1;
                    }
                    break;
                case "chat/closed":
                    this._chatButtonNotClicked = true;
                    if (this._chat.chatStatus !== 3) {
                        this._chat.chatStatus = 3;
                        thils.handleCloseChatClicked();
                    }
                    break;
                case "visInvite/visitorResponse":
                    if (inMessageObject.visitorInviteReaction === 0) {
                        this._chatState = State.kNone;
                        this._chatButtonNotClicked = true;
                    }
                    break;
                default:
                    throw new Error("ChatForm controller attempted to handle invalid request type");
            }
        }

        handleEvent(inEventMessage) {
            switch (inEventMessage.eventType) {
                case ChatMessages.chatEnabledRequestType:
                    thils.handleChatEnabled(inEventMessage);
                    break;
                case ChatMessages.chatButtonClickedRequestType:
                    this.handleChatButtonClicked(inEventMessage);
                    break;
                case ChatMessages.chatResurrectedRequestType:
                    this.handleChatResurrected(inEventMessage.data.chat, inEventMessage.data.userIdsInChat,
                        inEventMessage.data.userProfileThumbs, inEventMessage.data.defaultThumbnail,
                        inEventMessage.data.agentNameIdsInChat, inEventMessage.data.agentFirstNames,
                        inEventMessage.data.agentLastNames, inEventMessage.data.agentDisplayNames);

                    //MGMG TODO implement without event loop
                    // TC.EventMediator.mediateEvent(TC.chatStartedMessage());
                    break;
                case ChatMessages.serverErrorRequestType:
                    this._chatFormView.writeStatusMessage("An error has occurred, please contact the Solfa support team.");
                    this._chatFormView.writeStatusMessage("<a href='mailto:support@solfa.uk'>support@solfa.uk</a>");
                    this._chatFormView.disableChat();
                    this._chatFormView.hideUnpinButton();
                    this._chatFormView.setEndButtonToCloseButton();
                    break;
                case ChatMessages.visitorInviteAcceptedRequestType:
                    this.handleVisitorInviteClicked();
                    break;
                case ChatMessages.openingHoursViewRequestType:
                    let inMessageObject = inEventMessage;
                    this.handleBuildOpeningHoursForm(inMessageObject);
                    break;
                case ChatMessages.chatEndRequestType:
                    this.handleEndChatClicked();
                    break;
                case ChatMessages.showChatWindowRequestType:
                    this.handleShowChatWindow();
                    break;
                default:
                    throw new Error("ChatForm controller attempted to handle an invalid event type");
                    // break;
            }
        }

        chatID() { return this._chatId; }
        sessionID() { return this._sessionId; }

        // getEventTypes() { return this._eventTypes; }
        // this.getRequestTypes = function() { return mRequestTypes; }
        getChatId() {
            this._chatId = this._chatInviteId;
            return this._chatId;
        }
    }
    // ChatFormController.js
    // End Import

    class ChatWindowView {
        constructor(isExternalWindow, inBodyElement, inHeadElement) {
            this._isExternalWindow = isExternalWindow;
            this._bodyElement = inBodyElement;
            this._headElement = inHeadElement;
            //private data
            //Settings and HTML content.
            this._chatFormSettings                   = null;

            this._chatInput                          = null;
            this._sendChatMessageButton              = null;
            this._endChatButton                      = null;
            this._printChatButton                    = null;
            this._minimisedButton                    = null;
            this._exportChatButton                   = null;
            this._exportChatInput                    = null;
            this._unpinButton                        = null;
            this._notificationArea                   = null;

            //Top level of the chat window.
            this._chatFormRoot                       = null;
            //Chat form div
            this._chatFormDiv                        = null;
            //Chat form frame.
            this._chatFrame                          = null;

            this._sendButtonClickedCallback           = null;
            this._endButtonClickedCallback            = null;
            this._closeChatButtonClickedCallback      = null;
            this._joinButtonClickedCallback           = null;
            this._unpinButtonClickedCallback          = null;
            this._printButtonClickedCallback          = null;
            this._exportButtonClickedCallback         = null;
            this._visitorKeypressCallback             = null;
            this._minimisedClickedCallback            = null;
            this._expandedClickedCallback             = null;

            this._questionForm                        = null;
            this._userInput                           = null;
            this._queryInput                          = null;
            this._minimised                           = 0;

        }

        confirmChatMessageReceived() {
            this._chatInput.value = '';
            this.enableChat();
            this._chatInput.focus();
        }

        awaitMessageSentConfirmation() {
            this.disableChat();
        }

        onReconnection() {
            ETSChat.chatWindowView.enableChat();
            ETSChat.chatWindowView._chatInput.focus();
        }

        static onSendButtonClicked(e) {
            if (ETSChat.chatWindowView._sendButtonClickedCallback !== null) {
                ETSChat.chatWindowView._sendButtonClickedCallback(ETSChat.chatWindowView._chatInput.value);
            }
            if(ETSChat.chatWindowView._exportChatInput !== null) {
                ETSChat.chatWindowView._exportChatInput.value = '';
                ETSChat.chatWindowView._exportChatInput = null;
            }
        }

        escapeControlCharacters(str) {
            return str
                .replace(/[\\]/g, '\\\\')
                .replace(/[\"]/g, '\\\"')
                .replace(/[\/]/g, '\\/')
                .replace(/[\b]/g, '\\b')
                .replace(/[\f]/g, '\\f')
                .replace(/[\n]/g, '\\n')
                .replace(/[\r]/g, '\\r')
                .replace(/[\t]/g, '\\t')
                .replace(/[\v]/g, '\\v');
        }

        onVisitorKeypress(e) {
            if (ETSChat.chatWindowView._visitorKeypressCallback !== null)
                ETSChat.chatWindowView._visitorKeypressCallback(ETSChat.chatWindowView._chatInput.value);
        }

        onCloseChatButtonClicked(e) {
            if (ETSChat.chatWindowView._closeChatButtonClickedCallback !== null)
                ETSChat.chatWindowView._closeChatButtonClickedCallback();
        }

        onEndChatButtonClicked(e) {
            if (ETSChat.chatWindowView._endButtonClickedCallback !== null)
                ETSChat.chatWindowView._endButtonClickedCallback();
        }

        onJoinButtonClicked(e) {
            if (ETSChat.chatWindowView._joinButtonClickedCallback !== null)
                ETSChat.chatWindowView._joinButtonClickedCallback();
        }

        onUnpinButtonClicked() {
            if (ETSChat.chatWindowView._unpinButtonClickedCallback !== null)
                ETSChat.chatWindowView._unpinButtonClickedCallback();
        }

        onExportButtonClicked() {
            if (ETSChat.chatWindowView._exportButtonClickedCallback !== null)
                ETSChat.chatWindowView._exportButtonClickedCallback();
        }

        onPrintButtonClicked() {
            if(ETSChat.chatWindowView._printButtonClickedCallback !== null)
                ETSChat.chatWindowView._printButtonClickedCallback();
        }

        onMimiseButtonClicked() {
            ETSChat.chatWindowView._minimised = 1;
            for (let i = 0; i < ETSChat.chatWindowView._chatFrame.childNodes.length - 1; i++)
                ETSUtils.setStyleDisplay(this._chatFrame.childNodes[i], 0);

            let chatFrameRoot = ETSChat.chatWindowView._chatFormRoot.childNodes[0];
            chatFrameRoot.id       = "tcMainFrameMinimised";
            ETSChat.chatWindowView._chatFrame.onclick      = onExpandClicked;
            ETSChat.chatWindowView._minimisedClickedCallback();
        }

        onExpandClicked() {
            ETSChat.chatWindowView._minimised = 0;
            ETSChat.chatWindowView._chatFormRoot.childNodes[0].onclick = null;

            for(let i = 0; i < ETSChat.chatWindowView._chatFrame.childNodes.length -1; i++)
                ETSUtils.setStyleDisplay(ETSChat.chatWindowView._chatFrame.childNodes[i], 1);

            if (ETSUtils.checkVisitorIsOnMobile()) {
                if(ETSChat.chatWindowView._chatFormSettings.displayEmbeddedChatWindowOnMobile === true) {
                    ETSChat.chatWindowView._chatFormDiv.id                 = "tcMainFrameMobile";
                } else {
                    ETSChat.chatWindowView._chatFormDiv.id                 = "tcMainFrameMobileExternal";
                }
            } else {
                if (!ETSChat.chatWindowView._isExternalWindow) {
                    ETSChat.chatWindowView._chatFormDiv.id                 = "tcMainFrameEmbedded";
                } else {
                    ETSChat.chatWindowView._chatFormDiv.id                 = "tcMainFrameExternal";
                }
            }
            ETSChat.chatWindowView._chatFrame.scrollTop    = ETSChat.chatWindowView._chatFrame.scrollHeight;
            ETSChat.chatWindowView._expandedClickedCallback();
        }

        expandChatWindow() {
            this.onExpandClicked();
        }

        minimiseChatWindow() {
            this.onMimiseButtonClicked();
        }

        buildOutOfHoursForm(startTimesArray, endTimesArray, stateArray,
                                            companyEmail, emailAllowed, showBusinessHours) {

            this.hideMessageArea();
            this.hidePrintButton();
            this.hideExportButton();
            this.hideUnpinButton();
            this.setEndButtonToCloseButton();
            this.hideSendButton();

            //Create first two status messages.
            //var apologyMessage = document.createElement('p');
            //apologyMessage.id = 'tcBusinessHoursCaption1';
            //apologyMessage.id = 'tcFormOfflineChat';

            let introContainer = document.createElement('div');
            introContainer.id = 'tcFormOfflineChat';

            let apologyMessage = document.createElement('p');
            apologyMessage.id = 'tcIntroHeader1';

            let ourBusHoursMessage = document.createElement('p');
            ourBusHoursMessage.id = 'tcIntroHeader2';

            introContainer.appendChild(apologyMessage);
            introContainer.appendChild(ourBusHoursMessage);

            if(showBusinessHours == true) {

                let table = document.createElement('div');
                table.id = 'tcBusinessHoursTable';

                let tableHeaderGroup = document.createElement('div');
                tableHeaderGroup.id = 'tcBusinessHoursTableHeaderGroup';

                let headerItem = document.createElement('div');
                headerItem.className = 'tcBusinessHoursTableHeaderItem';
                tableHeaderGroup.appendChild(headerItem);

                let headerItemAm = document.createElement('div');
                headerItemAm.className = 'tcBusinessHoursTableHeaderItem';
                headerItemAm.innerHTML = "A.M.";
                tableHeaderGroup.appendChild(headerItemAm);

                let headerItemPm = document.createElement('div');
                headerItemPm.className = 'tcBusinessHoursTableHeaderItem';
                headerItemPm.innerHTML = "P.M.";
                tableHeaderGroup.appendChild(headerItemPm);

                this._chatFrame.appendChild(introContainer);
                //mChatFrame.appendChild(ourBusHoursMessage);
                table.appendChild(tableHeaderGroup);

                let days = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"];

                for(let i = 0; i < 7; i++) {
                    let businessHoursTimesGroup = document.createElement('div');
                    businessHoursTimesGroup.id = 'tcBusinessHoursTableRowGroup';

                    let businessHoursDay = document.createElement('div');
                    businessHoursDay.className = 'tcBusinessHoursTableRowItemDay';
                    businessHoursDay.innerHTML = days[i];

                    if(stateArray[i] == 0) {
                        let businessHoursClosed = document.createElement('div');
                        businessHoursClosed.className = 'tcBusinessHoursTableRowItemClosedDay';
                        businessHoursClosed.innerHTML = 'Closed';
                        businessHoursTimesGroup.appendChild(businessHoursDay);
                        businessHoursTimesGroup.appendChild(businessHoursClosed);
                    } else if (stateArray[i] == 2) {
                        let businessHoursOpen = document.createElement('div');
                        businessHoursOpen.className = 'tcBusinessHoursTableRowItemClosedDay';
                        businessHoursOpen.innerHTML = '24 Hours';
                        businessHoursTimesGroup.appendChild(businessHoursDay);
                        businessHoursTimesGroup.appendChild(businessHoursOpen);
                    } else {
                        let businessHoursStartTime = document.createElement('div');
                        businessHoursStartTime.className = 'tcBusinessHoursTableRowItemHour';
                        businessHoursStartTime.innerHTML = startTimesArray[i];

                        let businessHoursEndTime = document.createElement('div');
                        businessHoursEndTime.className = 'tcBusinessHoursTableRowItemHour';
                        businessHoursEndTime.innerHTML = endTimesArray[i];

                        businessHoursTimesGroup.appendChild(businessHoursDay);
                        businessHoursTimesGroup.appendChild(businessHoursStartTime);
                        businessHoursTimesGroup.appendChild(businessHoursEndTime);
                    }
                    table.appendChild(businessHoursTimesGroup);
                }

                this._chatFrame.appendChild(table);
            } else if (emailAllowed == true) {
                this._chatFormRoot.id                    = 'tcFormOfflineChat';

                let enterQuestionLabel            = document.createElement('p');
                enterQuestionLabel.id             = 'tcBusinessHoursQuestionLabel';
                enterQuestionLabel.innerHTML      = '<br > If you would like to email the company with a query, please enter your email and query below and we will get back to you as soon as possible.';
                this._chatFrame.appendChild(enterQuestionLabel);

                let formShell = document.createElement('div');
                formShell.id = 'tcFormEmailOfflineChat';
                let questionForm = document.createElement('div');
                questionForm.id        = 'tcChatForm';
                questionForm.className = 'tcChatForm';

                let questionBox = document.createElement('div');
                questionBox.className = 'tcQuestionBox';

                let emailEditQuestion = document.createElement('div');
                emailEditQuestion.className = 'tcEditQuestion';
                emailEditQuestion.innerHTML = 'Email';

                let emailEditAnswer         = document.createElement('div');
                emailEditAnswer.className   = 'tcEditAnswer';

                /*userInput = document.createElement('div');
                this._userInput.id        = 'tcOfflineEmailAnswer';
                this._userInput.className = 'tcEdit';
                this._userInput.contentEditable = true;
                this._userInput.placeholder = 'Type Email Here...';*/

                this._userInput = document.createElement('input');
                this._userInput.id        = 'tcOfflineEmailAnswer';
                this._userInput.className = 'tcEdit';
                this._userInput.placeholder = 'Type Email Here...';

                emailEditAnswer.appendChild(userInput);
                questionBox.appendChild(emailEditQuestion);
                questionBox.appendChild(emailEditAnswer);

                let questionBox2 = document.createElement('div');
                questionBox2.className = 'tcQuestionBox';

                let queryEditQuestion = document.createElement('div');
                queryEditQuestion.className = 'tcEditQuestion';
                queryEditQuestion.innerHTML = 'Question To Company';

                let queryEditAnswer = document.createElement('div');
                queryEditAnswer.className = 'tcEditAnswer';

                /*queryInput = document.createElement('div');
                this._queryInput.id        = 'tcOfflineQuestionAnswer';
                this._queryInput.className = 'tcEdit';
                this._queryInput.contentEditable = true;
                this._queryInput.placeholder = 'Type Your Query...';*/

                this._queryInput = document.createElement('TEXTAREA');
                this._queryInput.id        = 'tcOfflineQuestionAnswer';
                this._queryInput.className = 'tcTextArea';
                this._queryInput.placeholder = 'Write a message here...';


                queryEditAnswer.appendChild(queryInput);
                questionBox2.appendChild(queryEditQuestion);
                questionBox2.appendChild(queryEditAnswer);

                questionForm.appendChild(questionBox);
                questionForm.appendChild(questionBox2);
                formShell.appendChild(questionForm);
                this._chatFrame.appendChild(formShell);

                this.enableSendButton();
                this.showSendButton();
            }
            if(showBusinessHours == false) {
                introContainer.removeChild(ourBusHoursMessage);
                this._chatFrame.appendChild(introContainer);
            }

            this._chatFormDiv.insertBefore(this._chatFrame, this._chatFormDiv.childNodes[2]);

            this.showChatWindow();
        }

        buildChatFormQuestion(inQuestionType, inQuestion, inAnswers, inQNum, inOptionalId) {
            //TODO::REfactor and implement id's so that data can be pulled from each question built.
            let divClassQuestionBox         = document.createElement('div');
            divClassQuestionBox.className   = 'tcQuestionBox';
            let divClassQuestion        = document.createElement('div');

            if (inOptionalId !== undefined && inOptionalId !== "")
                divClassQuestion.id = inOptionalId;


            if (inQuestionType == OptionType.kPlainText) {
                divClassQuestion.className      = 'tcEditQuestion';
                divClassQuestion.innerHTML      = inQuestion;


                let divClassInput               = document.createElement('div');
                divClassInput.className         = 'tcEditAnswer';

                if(inOptionalId === "tcQFeedback" || inOptionalId === undefined) {
                    let textArea = document.createElement('TEXTAREA');
                    textArea.className = "tcTextArea";
                    textArea.placeholder = "Type Answer Here..";

                    divClassQuestionBox.appendChild(divClassQuestion);
                    divClassInput.appendChild(textArea);
                    divClassQuestionBox.appendChild(divClassInput);
                } else {

                    let inputArea                   = document.createElement('input');
                    inputArea.className             = 'tcEdit';
                    inputArea.placeholder           = 'Type Here...';

                    //Changed as we had a styling issue with our css(box's with no style would be invis).

                    //var divClassEdit                = document.createElement('div');
                    //divClassEdit.className          = 'tcEdit';
                    //divClassEdit.contentEditable    = true

                    //divClassInput.appendChild(divClassEdit);
                    //divClassQuestionBox.appendChild(divClassInput);

                    divClassQuestionBox.appendChild(divClassQuestion);
                    divClassInput.appendChild(inputArea);
                    divClassQuestionBox.appendChild(divClassInput);
                }



            } else if (inQuestionType == OptionType.kRadioButton) {
                divClassQuestion.className  = 'tcRadioQuestion';
                divClassQuestion.innerHTML  = inQuestion;

                divClassQuestionBox.appendChild(divClassQuestion);

                if (inAnswers !== '' && inAnswers !== undefined && inAnswers !== null) {
                    if (inAnswers.length > 0) {
                        for(let i = 0; i < inAnswers.length; i ++) {
                            let labelDiv        = document.createElement('label');
                            labelDiv.htmlFor    ='tcQ' +inQNum + "A" + i;

                            let inputDiv        = document.createElement('input');
                            inputDiv.id         = 'tcQ' + inQNum + "A" + i;
                            inputDiv.type       ='radio';
                            inputDiv.name       = 'tcQ' + inQNum;
                            inputDiv.value      = 'tcQ' + inQNum + "A" + i;

                            let spanDiv         = document.createElement('span');
                            spanDiv.innerHTML   = inAnswers[i];

                            labelDiv.appendChild(inputDiv);
                            labelDiv.appendChild(spanDiv);

                            divClassQuestionBox.appendChild(labelDiv);
                        }
                    }
                }
            } else if (inQuestionType == OptionType.kCheckBox) {
                let divClassQuestion        = document.createElement('div');
                divClassQuestion.className  = 'tcCheckQuestion';
                divClassQuestion.innerHTML  = inQuestion;

                divClassQuestionBox.appendChild(divClassQuestion);

                if (inAnswers !== '' && inAnswers !== undefined && inAnswers !== null) {
                    if (inAnswers.length > 0) {
                        for(let i = 0; i < inAnswers.length; i ++) {
                            let labelDiv        = document.createElement('label');
                            labelDiv.htmlFor    ='tcQ' +inQNum + "A" + i;

                            let inputDiv        = document.createElement('input');
                            inputDiv.id         = 'tcQ' + inQNum + "A" + i;
                            inputDiv.type       ='checkbox';
                            inputDiv.name       = 'tcQ' + inQNum;
                            inputDiv.value      = 'tcQ' + inQNum + "A" + i;

                            let spanDiv         = document.createElement('span');
                            spanDiv.innerHTML   = inAnswers[i];

                            labelDiv.appendChild(inputDiv);
                            labelDiv.appendChild(spanDiv);

                            divClassQuestionBox.appendChild(labelDiv);
                        }
                    }
                }
            }
            return divClassQuestionBox;
        }


        setSendButtonClickedCallback(inCallback) {
            this._sendButtonClickedCallback           = inCallback;
        }

        setEndButtonClickedCallback(inCallback) {
            this._endButtonClickedCallback            = inCallback;
        }

        setCloseChatButtonClickedCallback(inCallback) {
            this._closeChatButtonClickedCallback      = inCallback;
        }

        setJoinButtonClickedCallback(inCallback) {
            this._joinButtonClickedCallback = inCallback;
        }

        setUnpinButtonClickedCallback(inCallback) {
            this._unpinButtonClickedCallback          = inCallback;
        }

        setPrintButtonClickedCallback(inCallback) {
            this._printButtonClickedCallback          = inCallback;
        }

        setExportButtonClickedCallback(inCallback) {
            this._exportButtonClickedCallback         = inCallback;
        }

        setVisitorKeyPressCallback(inCallback) {
            this._visitorKeypressCallback             = inCallback;
        }

        setMinimisedClickedCallback(inCallback) {
            this._minimisedClickedCallback = inCallback;
        }

        setExpandedClickedCallback(inCallback) {
            this._expandedClickedCallback = inCallback;
        }

        setUISettings(inChatFormSettings) {

            this._chatFormSettings                   = inChatFormSettings;

            this._chatFormRoot                       = document.createElement('div');
            this._chatFormRoot.id                    = 'tcFormInChat';

            this._chatFormDiv                        = document.createElement('div');


            if (ETSUtils.checkVisitorIsOnMobile()) {
                if(this._chatFormSettings.displayEmbeddedChatWindowOnMobile === true) {
                    this._chatFormDiv.id                 = "tcMainFrameMobile";

                    //viewport = document.querySelector("meta[name=viewport]")
                    //viewport.setAttribute('content', 'width=device-width, initial-scale=1, maximum-scale=1');

                } else {
                    this._chatFormDiv.id                 = "tcMainFrameMobileExternal";
                }
            } else {
                if (!this._isExternalWindow) {
                    this._chatFormDiv.id                 = "tcMainFrameEmbedded";
                } else {
                    this._chatFormDiv.id                 = "tcMainFrameExternal";
                }
            }

            // MGMG TODO -- fix the innerHTML content
            // this._chatFormDiv.innerHTML                  = "{{{$file-remove-nl script/TonicChatMainFrame.html}}}";
            this._chatFormDiv.innerHTML = '<div class=\'tcBranding\' id=\'tcBrandingTop\'></div>\n' +
                '<div id=\'tcControlPanel\'>\n' +
                '<button class=\'tcControlButton\' id=\'tcCloseWindowButton\'></button>\n' +
                '<button class=\'tcControlButton\' id=\'tcMinimiseWindowButton\'></button>\n' +
                '<button class=\'tcControlButton\' id=\'tcUnpinButton\'></button>\n' +
                '<button class=\'tcControlButton\' id=\'tcPrintTranscriptButton\'></button>\n' +
                '<button class=\'tcControlButton\' id=\'tcEmailTranscriptButton\'></button>      \n' +
                '</div>\n' +
                '<div id=\'tcMainChatFrame\'>\n' +
                '</div>\n' +
                '<div id=\'tcfooter\'>\n' +
                '<div id=\'tcBottomControlPanel\'>\n' +
                '<div id=\'tcNotification\'></div>\n' +
                '<div id=\'tcNewMessageBoxDiv\'> <textarea id=\'tcNewMessageArea\' placeholder=\'Type message here...\'></textarea></div>\n' +
                '<div id=\'tcSendButtonDiv\'><button class=\'tcControlButton\' id=\'tcSendButton\'></button></div>\n' +
                '</div>\n' +
                '<div class=\'tcBranding\' id=\'tcBrandingBottom\'>ETS Chat Powered By Engage To Sell LLC.</div>\n' +
                '</div>';

            if(this._chatFormSettings.headerBranding !== "" && this._chatFormSettings.headerBranding !== undefined) {
                this._chatFormDiv.childNodes[0].innerHTML = '<img src = "data:image/' + "png" + ';base64,' + this._chatFormSettings.headerBranding + '"></img>';
            }

            if(this._chatFormSettings.footerBranding !== "" && this._chatFormSettings.footerBranding !== undefined) {
                let children = this._chatFormDiv.childNodes;
                for(let i = 0; i < children.length; i++) {
                    if(children[i].id === 'tcfooter') {
                        let inFooter = children[i].childNodes;
                        for(let j = 0; inFooter.length; j++) {
                            if(inFooter[j].id === 'tcBrandingBottom') {
                                inFooter[j].innerHTML = '<img src = "data:image/' + "png" + ';base64,' + mChatFormSettings.footerBranding + '"></img>';
                                break;
                            }
                        }
                    }
                }
            }

            let tcMainFrame                     = "";
            let chatChildNodes                  = this._chatFormDiv.childNodes;

            let foundTcMainChatFrame            = 0;
            let foundControlPanel               = 0;
            let foundFooter                     = 0;

            let foundCloseWindowButton          = 0;
            let foundUnpinButton                = 0;
            let foundMinimisedButton            = 0;
            let foundPrintButton                = 0;
            let foundExportButton               = 0;

            let foundSendButton                 = 0;
            let foundMessageArea                = 0;
            let foundNotificationArea           = 0;

            for(let i = 0; i < chatChildNodes.length; i++) {
                if (chatChildNodes[i].id === 'tcControlPanel') {
                    foundControlPanel               = 1;

                    let controlPanelChildNodes      = chatChildNodes[i].childNodes;

                    for(let j = 0; j < controlPanelChildNodes.length; j++) {

                        let controlPanelNode        = controlPanelChildNodes[j];

                        if (controlPanelNode.id === 'tcCloseWindowButton') {

                            this._endChatButton          = controlPanelNode;

                            foundCloseWindowButton  = 1;
                        } else if (controlPanelNode.id === 'tcUnpinButton') {

                            this._unpinButton            = controlPanelNode;
                            foundUnpinButton        = 1;

                        } else if (controlPanelNode.id === 'tcPrintTranscriptButton') {

                            this._printChatButton        = controlPanelNode;
                            foundPrintButton        = 1;

                        } else if (controlPanelNode.id === 'tcEmailTranscriptButton') {
                            this._exportChatButton       = controlPanelNode;
                            foundExportButton       = 1;
                        } else if (controlPanelNode.id === 'tcMinimiseWindowButton') {
                            this._minimisedButton        = controlPanelNode;
                            foundMinimisedButton    = 1;
                        }

                        if (foundCloseWindowButton && foundUnpinButton && foundPrintButton
                            && foundExportButton && foundMinimisedButton)
                            break;
                    }

                    if (!foundCloseWindowButton)
                        throw new Error("ChatFormView could not find the tcCloseWindowButton element");

                    if (!foundUnpinButton)
                        throw new Error("ChatFormView could not find the tcUnpinButton element");

                    if (!foundPrintButton)
                        throw new Error("ChatFormView could not find the tcPrintTranscriptButton element");

                    if (!foundExportButton)
                        throw new Error("ChatFormView could not find the tcEmailTranscriptButton element");

                    if (!foundMinimisedButton)
                        throw new Error("ChatFormView could not find the tcMinimiseWindowButton element");

                    if (this._isExternalWindow)
                        chatChildNodes[i].removeChild(mUnpinButton);

                } else if (chatChildNodes[i].id === 'tcMainChatFrame') {
                    this._chatFrame                      = chatChildNodes[i];
                    foundTcMainChatFrame            = 1;
                } else if (chatChildNodes[i].id === 'tcfooter') {
                    // let messagePanelChildNodes      = chatChildNodes[i].childNodes[0].childNodes;
                    let messagePanelChildNodes      = chatChildNodes[i].childNodes[1].childNodes;
                    foundFooter                     = 1;
                    for(let j = 0; j < messagePanelChildNodes.length; j++) {
                        if (messagePanelChildNodes[j].id === 'tcNotification') {
                            this._notificationArea       = messagePanelChildNodes[j];
                            foundNotificationArea   = 1;
                        } else if (messagePanelChildNodes[j].id === 'tcNewMessageBoxDiv') {
                            for(let k = 0; k < messagePanelChildNodes[j].childNodes.length; k++) {
                                let messagePanelChildNode = messagePanelChildNodes[j].childNodes[k];
                                if (messagePanelChildNode.id === 'tcNewMessageArea') {
                                    this._chatInput = messagePanelChildNode;
                                    foundMessageArea = 1;
                                    break;
                                }
                            }
                            /*mChatInput              = messagePanelChildNodes[j].childNodes[0].nextSibling;
                            alert(mChatInput);
                            foundMessageArea        = 1;*/
                        } else if (messagePanelChildNodes[j].id === 'tcSendButtonDiv') {
                            this._sendChatMessageButton  = messagePanelChildNodes[j].childNodes[0];
                            foundSendButton         = 1;
                        }
                    }
                }

                if (foundTcMainChatFrame && foundControlPanel && foundMessageArea)
                    break;
            }

            if (!foundTcMainChatFrame)
                throw new Error("ChatFormView could not find the tcMainFrame element");

            if (!foundSendButton)
                throw new Error("PreChatFormView could not find the send button element");

            if (!foundMessageArea)
                throw new Error("PreChatFormView could not find the tcNewMessageArea element");

            if (!foundNotificationArea)
                throw new Error("ChatFormView could not find the tcNotificationArea element");

            this._chatFormRoot.appendChild(this._chatFormDiv);

            ETSUtils.setStyleDisplay(this._chatFormRoot, 0);

            this._bodyElement.appendChild(this._chatFormRoot);

            this._chatInput.onkeypress = function(e) {
                if (!e)
                    e = window.event;
                let keyCode = e.keyCode || e.which;
                if (keyCode === 13) {
                    if (e.shiftKey === true)
                        this._chatInput.innerHTML += '\n';
                    else {
                        e.preventDefault();
                        ChatWindowView.onSendButtonClicked();
                    }
                }
            };

            this._unpinButton.onclick = this.onUnpinButtonClicked;
            this._exportChatButton.onclick = this.onExportButtonClicked;
            this._printChatButton.onclick = this.onPrintButtonClicked;
            this._minimisedButton.onclick = this.onMimiseButtonClicked;
        }

        insertBreakInChatForm() {
            this._chatFrame.innerHTML += "<br />";
        }

        createInChatForm(inChatFormId, inChatFormData) {

            let chatFormDivParent = document.createElement('div');
            chatFormDivParent.id = inChatFormId;

            if (inChatFormData.chatFormIntroOne !== '') {
                let introHeaderOneDiv = document.createElement('div');
                introHeaderOneDiv.id = 'tcChatFormIntroHeader1';
                introHeaderOneDiv.innerHTML = inChatFormData.chatFormIntroOne;
                chatFormDivParent.appendChild(introHeaderOneDiv);
            }

            if (inChatFormData.chatFormIntroTwo !== '') {
                let introHeaderTwoDiv = document.createElement('div');
                introHeaderTwoDiv.id = 'tcChatFormIntroHeader2';
                introHeaderTwoDiv.innerHTML = inChatFormData.chatFormIntroTwo;
                chatFormDivParent.appendChild(introHeaderTwoDiv);
            }

            let inChatFormDiv = document.createElement('div');
            inChatFormDiv.id = 'tcChatForm';

            for(let i = 0; i < inChatFormData.chatQuestions.length; i++) {
                let question = inChatFormData.chatQuestions[i];

                let questionDiv = buildChatFormQuestion(question.optionType,
                    question.question, question.options, i, question.id);

                inChatFormDiv.appendChild(questionDiv);
            }
            chatFormDivParent.appendChild(inChatFormDiv);
            this._chatFrame.appendChild(chatFormDivParent);

            if(chatFormDivParent.id === "tcExportForm") {
                this._exportChatInput = document.getElementsByClassName("tcEdit")[0];
                this._exportChatInput.focus();

                this._exportChatInput.onkeypress = function(e) {
                    if (!e)
                        e = window.event;
                    let keyCode = e.keyCode || e.which;
                    if (keyCode === 13) {
                        e.preventDefault();
                        ChatWindowView.onSendButtonClicked();
                    }
                }
                //document.getElementsByClassName("tcEdit")[0].focus();
            }
        }

        getInChatForm(inChatFormId) {
            let chatForm = null;
            let chatFrameChildren = this._chatFrame.childNodes;

            for(let i = 0; i < chatFrameChildren.length; i++) {
                let childNode = chatFrameChildren[i];
                if (childNode.id === inChatFormId) {
                    for(let j = 0; j < childNode.childNodes.length; j++) {
                        let chatFormChildNode = childNode.childNodes[j];
                        if (chatFormChildNode.id === "tcChatForm") {
                            chatForm = chatFormChildNode;
                            break;
                        }
                    }
                }
            }

            return chatForm;
        }

        removeInChatForm(inChatFormId) {
            let chatForm = null;

            if (inChatFormId === "" || inChatFormId === undefined)
                throw new Error("Failed to remove chat form as the chat form id is null or undefined");
            else {
                let chatFrameChildren = this._chatFrame.childNodes;
                for(let i = 0; i < chatFrameChildren.length; i++) {
                    let childNode = chatFrameChildren[i];
                    if (childNode.id === inChatFormId) {

                        for(let j = 0; j < childNode.childNodes.length; j++) {
                            let chatFormChildNode = childNode.childNodes[j];
                            if (chatFormChildNode.id === "tcChatForm") {
                                chatForm = chatFormChildNode;
                                break;
                            }
                        }

                        this._chatFrame.removeChild(childNode);
                        break;
                    }
                }
            }

            return chatForm;
        }

        writeStatusMessage(inMessage, id) {
            /*mChatFrame.innerHTML += "<div"

            if (id !== undefined || id !== "")
                mChatFrame.innerHTML += " id=" + id;

            mChatFrame.innerHTML += " class = \"tcChatNoficiation\">" + inMessage + "</div>";
            mChatFrame.scrollTop  = mChatFrame.scrollHeight;*/

            let statusMessage = document.createElement('div');

            if (id !== undefined && id !== "") {
                statusMessage.id = id;
            }
            statusMessage.className = "tcChatNotification";
            statusMessage.innerHTML += inMessage;

            this._chatFrame.appendChild(statusMessage);
            this._chatFrame.scrollTop = this._chatFrame.scrollHeight;
        }

        setNotification(inMessage) {
            this._notificationArea.innerHTML = inMessage;
        }

        writeMessage(inIsOutgoingMessage, inName, inTimestamp, inMessage, inAgentThumbnail) {

            let   groupClassPrefix       = inIsOutgoingMessage ? "OutgoingContainer" : "IncomingContainer";
            let   from                   = inName;
            let   utcSeconds             = (!inTimestamp || inTimestamp === '') ? (new Date).getTime() : inTimestamp;
            let   date                   = new Date(utcSeconds);


            let timestampFormated = [
                (date.getMonth() + 1) + "/" +
                date.getDate() + "/" +
                date.getFullYear() + ": " +
                date.getHours().toString().padLeft("0",2) + ":" +
                date.getMinutes().toString().padLeft("0",2) + ":" +
                date.getSeconds().toString().padLeft("0",2)
            ];


            let messageContainer = document.createElement('div');
            messageContainer.className = "tcMessage" + groupClassPrefix;

            let photoDiv = document.createElement('div');
            photoDiv.className = 'tcPhoto';

            if (inAgentThumbnail !== null && inAgentThumbnail !== undefined
                && inAgentThumbnail !== "") {
                let agentThumbnail = '<img src ="data:image/png;base64,' + inAgentThumbnail + '"></img>';
                photoDiv.innerHTML = agentThumbnail;
            }

            messageContainer.appendChild(photoDiv);

            let messageGroupDiv = document.createElement('div');
            messageGroupDiv.className = 'tcMessageGroup';

            let messageHeader = document.createElement('div');

            messageHeader.className = 'tcHeader';
            messageHeader.innerHTML = from + ", " + timestampFormated;

            messageGroupDiv.appendChild(messageHeader);

            let messageDiv = document.createElement('div');
            messageDiv.className = "tcMessage";
            messageDiv.innerHTML = inMessage;

            messageGroupDiv.appendChild(messageDiv);

            messageContainer.appendChild(messageGroupDiv);

            if (this._minimised) {
                ETSUtils.setStyleDisplay(this._chatFrame.lastChild, 0);
            }

            this._chatFrame.appendChild(messageContainer);
            this._chatFrame.scrollTop         = this._chatFrame.scrollHeight;
        }

        showChatWindow() {
            ETSUtils.setStyleDisplay(this._chatFormRoot, 1);
        }

        createIntroHeader() {
            let introContainer = document.createElement('div');
            introContainer.id = 'tcIntroContainer';
            let introHeaderOne = document.createElement('div');
            introHeaderOne.id = 'tcIntroHeader1';
            let introHeaderTwo = document.createElement('div');
            introHeaderTwo.id = 'tcIntroHeader2';
            introContainer.appendChild(introHeaderOne);
            introContainer.appendChild(introHeaderTwo);
            this._chatFrame.appendChild(introContainer);
        }

        hideChatWindow() {
            ETSUtils.setStyleDisplay(this._chatFormRoot, 0);
        }

        clearChatWindow() {
            this._chatFrame.innerHTML = '';
        }

        hideUnpinButton() {
            ETSUtils.setStyleDisplay(this._unpinButton, 0);
        }

        showUnpinButton() {
            ETSUtils.setStyleDisplay(this._unpinButton, 1);
        }

        showMessageArea() {
            ETSUtils.setStyleDisplay(this._chatInput, 1);
        }

        hideMessageArea() {
            ETSUtils.setStyleDisplay(this._chatInput, 0);
        }

        showExportButton() {
            ETSUtils.setStyleDisplay(this._exportChatButton, 1);
        }

        hideExportButton() {
            ETSUtils.setStyleDisplay(this._exportChatButton, 0);

        }

        showPrintButton() {
            ETSUtils.setStyleDisplay(this._printChatButton, 1);
        }

        hidePrintButton() {
            ETSUtils.setStyleDisplay(this._printChatButton, 0);
        }

        showSendButton() {
            ETSUtils.setStyleDisplay(this._sendChatMessageButton, 1);
        }

        hideSendButton() {
            ETSUtils.setStyleDisplay(this._sendChatMessageButton, 0);
        }

        setEndButtonToCloseButton() {
            ETSChat.chatWindowView._endChatButton.value    = "Close";
            ETSChat.chatWindowView._endChatButton.onclick  = ETSChat.chatWindowView.onCloseChatButtonClicked;
        }

        setEndButtonToJoinButton() {
            this._endChatButton.value                = "Join Chat";
            this._endChatButton.onclick              = this.onJoinButtonClicked;
        }

        setJoinButtonToEndButton() {
            this._endChatButton.value                = "End Chat";
            this._endChatButton.onclick              = this.onEndChatButtonClicked;
        }

        setJoinButtonToCloseButton() {
            this._endChatButton.value                = "Close";
            this.endChatButton.onclick              = this.onCloseChatButtonClicked;
        }

        setCloseButtonToEndButton() {
            this._endChatButton.value                = "End Chat";
            this._endChatButton.onclick              = this.onEndChatButtonClicked;
        }

        enableSendButton() {
            this._sendChatMessageButton.onclick = ChatWindowView.onSendButtonClicked;
            this._sendChatMessageButton.disabled = false;
        }

        disableSendButton() {
            this._sendChatMessageButton.onclick = null;
            this._sendChatMessageButton.disabled = true;
        }

        disableEndButton() {
            this._endChatButton.onclick = null;
            this._endChatButton.disabled = true;
        }

        disableChat()  {
            this._sendChatMessageButton.onclick      = null;
            this._chatInput.onkeyup                  = null;
            this._endChatButton.onclick              = null;

            this._chatInput.disabled                 = true;
            this._sendChatMessageButton.disabled     = true;
            this._endChatButton.onclick              = this.onCloseChatButtonClicked;
        }

        changeEndButtonToClose() {
            this._endChatButton.onclick              = this.onCloseChatButtonClicked;
        }

        enableChat() {
            if(this._sendChatMessageButton === null) return;
            if(this._chatInput === null) return;
            if(this._endChatButton === null) return;
            if(this._sendChatMessageButton === null) return;

            this._sendChatMessageButton.onclick      = ChatWindowView.onSendButtonClicked;
            this._chatInput.onkeyup                  = ChatWindowView.onVisitorKeypress;
            this._endChatButton.onclick              = ChatWindowView.onEndChatButtonClicked;

            this._chatInput.disabled                 = false;
            this._sendChatMessageButton.disabled     = false;
        }
    }


    class AlignmentManager {

        static get kTopLeft()             { return 17; }
        static get kTopLeftCentre()       { return  18; }
        static get kTopCentre()           { return  16; }
        static get kTopRightCentre()      { return  20; }
        static get kTopRight()            { return  24; }

        static get kLeftFromTop()         { return 33; }
        static get kLeftCentre()          { return 1; }
        static get kLeftFromBottom()      { return 65; }

        static get kRightFromTop()        { return 40; }
        static get kRightCentre()         { return 8; }
        static get kRightFromBottom()     { return  72; }

        static get kBottomLeft()          { return  129; }
        static get kBottomLeftCentre()    { return  130; }
        static get kBottomCentre()        { return  128; }
        static get kBottomRightCentre()   { return  132; }
        static get kBottomRight()         { return  136; }

        constructor(inBodyElement, inHeadElement) {
            this._bodyElement = inBodyElement;
            this._headElement = inHeadElement;
            /*
            var centre = document.createElement('div');
            centre.className = "tcButtonsAlignmentX2Y2";
            */
            this._kAlignmentDivs = [];

        }



        addWidget(widget, newAlignmentNumber, oldAlignmentNumber) {
            if(newAlignmentNumber == oldAlignmentNumber)
                return;

            let newAlignmentName = null;
            let oldAlignmentName = null;

            switch(newAlignmentNumber) {
                case AlignmentManager.kTopLeft:
                    newAlignmentName = "tcButtonsAlignmentX0Y0";
                    break;
                case AlignmentManager.kTopLeftCentre:
                    newAlignmentName = "tcButtonsAlignmentX1Y0";
                    break;
                case AlignmentManager.kTopCentre:
                    newAlignmentName = "tcButtonsAlignmentX2Y0";
                    break;
                case AlignmentManager.kTopRightCentre:
                    newAlignmentName = "tcButtonsAlignmentX3Y0";
                    break;
                case AlignmentManager.kTopRight:
                    newAlignmentName = "tcButtonsAlignmentX4Y0";
                    break;
                case AlignmentManager.kLeftFromTop:
                    newAlignmentName = "tcButtonsAlignmentX0Y1";
                    break;
                case AlignmentManager.kLeftCentre:
                    newAlignmentName = "tcButtonsAlignmentX0Y2";
                    break;
                case AlignmentManager.kLeftFromBottom:
                    newAlignmentName = "tcButtonsAlignmentX0Y3";
                    break;
                case AlignmentManager.kRightFromTop:
                    newAlignmentName = "tcButtonsAlignmentX4Y1";
                    break;
                case AlignmentManager.kRightCentre:
                    newAlignmentName = "tcButtonsAlignmentX4Y2";
                    break;
                case AlignmentManager.kRightFromBottom:
                    newAlignmentName = "tcButtonsAlignmentX4Y3";
                    break;
                case AlignmentManager.kBottomLeft:
                    newAlignmentName = "tcButtonsAlignmentX0Y4";
                    break;
                case AlignmentManager.kBottomLeftCentre:
                    newAlignmentName = "tcButtonsAlignmentX1Y4";
                    break;
                case AlignmentManager.kBottomCentre:
                    newAlignmentName = "tcButtonsAlignmentX2Y4";
                    break;
                case AlignmentManager.kBottomRightCentre:
                    newAlignmentName = "tcButtonsAlignmentX3Y4";
                    break;
                case AlignmentManager.kBottomRight:
                    newAlignmentName = "tcButtonsAlignmentX4Y4";
                    break;
                default:
                    return;
            }

            this.insertWidget (widget, newAlignmentName);

            if (oldAlignmentNumber !== 0) {
                switch(oldAlignmentNumber) {
                    case AlignmentManager.kTopLeft:
                        oldAlignmentName = "tcButtonsAlignmentX0Y0";
                        break;
                    case AlignmentManager.kTopLeftCentre:
                        oldAlignmentName = "tcButtonsAlignmentX1Y0";
                        break;
                    case AlignmentManager.kTopCentre:
                        oldAlignmentName = "tcButtonsAlignmentX2Y0";
                        break;
                    case AlignmentManager.kTopRightCentre:
                        oldAlignmentName = "tcButtonsAlignmentX3Y0";
                        break;
                    case AlignmentManager.kTopRight:
                        oldAlignmentName = "tcButtonsAlignmentX4Y0";
                        break;
                    case AlignmentManager.kLeftFromTop:
                        oldAlignmentName = "tcButtonsAlignmentX0Y1";
                        break;
                    case AlignmentManager.kLeftCentre:
                        oldAlignmentName = "tcButtonsAlignmentX0Y2";
                        break;
                    case AlignmentManager.kLeftFromBottom:
                        oldAlignmentName = "tcButtonsAlignmentX0Y3";
                        break;
                    case AlignmentManager.kRightFromTop:
                        oldAlignmentName = "tcButtonsAlignmentX4Y1";
                        break;
                    case AlignmentManager.kRightCentre:
                        oldAlignmentName = "tcButtonsAlignmentX4Y2";
                        break;
                    case AlignmentManager.kRightFromBottom:
                        oldAlignmentName = "tcButtonsAlignmentX4Y3";
                        break;
                    case AlignmentManager.kBottomLeft:
                        oldAlignmentName = "tcButtonsAlignmentX0Y4";
                        break;
                    case AlignmentManager.kBottomLeftCentre:
                        oldAlignmentName = "tcButtonsAlignmentX1Y4";
                        break;
                    case AlignmentManager.kBottomCentre:
                        oldAlignmentName = "tcButtonsAlignmentX2Y4";
                        break;
                    case AlignmentManager.kBottomRightCentre:
                        oldAlignmentName = "tcButtonsAlignmentX3Y4";
                        break;
                    case AlignmentManager.kBottomRight:
                        oldAlignmentName = "tcButtonsAlignmentX4Y4";
                        break;
                    default:
                        return;
                }

                this.removeWidget (widget, oldAlignmentName);
            }
        }

        insertWidget(widget, alignmentName) {
            let alignmentDiv = null;

            for(let i = 0; i < this._kAlignmentDivs.length; i++) {
                var buttonAlignment = this._kAlignmentDivs[i];

                if (buttonAlignment.className == alignmentName) {
                    alignmentDiv = buttonAlignment;
                }
            }

            if(alignmentDiv) {
                alignmentDiv.appendChild(widget);
            } else {
                const newDiv = document.createElement('div');
                newDiv.className = alignmentName;
                newDiv.appendChild(widget);

                this._bodyElement.appendChild(newDiv);

                this._kAlignmentDivs[this._kAlignmentDivs.length] = newDiv;
            }
        }

        removeWidget(widget, alignmentName) {
            for(let i = 0; i < this._kAlignmentDivs.length; i++) {
                let buttonAlignment = this._kAlignmentDivs[i];

                if (buttonAlignment.className === alignmentName) {
                    let buttonAlignmentChilds = buttonAlignment.childNodes;

                    if (buttonAlignmentChilds.length === 0) {
                        this._bodyElement.removeChild(buttonAlignment);
                        this._kAlignmentDivs.splice(i, 1);
                    }
                }
            }
        }
    }

    var OptionType = {
        kPlainText   : 0x01,
        kRadioButton : 0x02,
        kCheckBox : 0x03
    };

    class ETSTimer {

        constructor(callback, period, enabled) {
            this._callback = callback;
            this._period = period || 0;
            this._enabled = enabled || false;

            this.restartTimer();
        }

        setPeriod(period) {
            if (period != this._period) {
                this._period = period;
                this.restartTimer();
            }
        }

        getPeriod () {
            return this.period;
        }

        setEnabled(enabled) {
            if (enabled != this._enabled) {
                this._enabled = enabled;
                this.restartTimer();
            }
        };

        getEnabled() {
            return this._enabled;
        }

        restartTimer() {
            if (this._id !== undefined) {
                clearInterval(this._id);
                delete this._id;
            }
            if (this._enabled) {
                this._id = setInterval(this._callback, this._period);
            }
        }
    }


    class ETSUtils {
        constructor() {

        }

        static createChatFormQuestion() {
            let chatFormQuestion = {
                id : "",
                optionType : -1,
                question : "",
                options : []
            };

            return chatFormQuestion;
        };

        static createChatFormObject() {
            let chatFormObject = {
                chatFormIntroOne : "",
                chatFormIntroTwo : "",
                chatQuestions: []
            };

            return chatFormObject;
        };

        static checkVisitorIsOnMobile() {
            var isMobile = false;
            (function(a){if(/(android|bb\d+|meego).+mobile|avantgo|bada\/|blackberry|blazer|compal|elaine|fennec|hiptop|iemobile|ip(hone|od)|iris|kindle|lge |maemo|midp|mmp|mobile.+firefox|netfront|opera m(ob|in)i|palm( os)?|phone|p(ixi|re)\/|plucker|pocket|psp|series(4|6)0|symbian|treo|up\.(browser|link)|vodafone|wap|windows ce|xda|xiino|android|ipad|playbook|silk/i.test(a)||/1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|\-[a-w])|libw|lynx|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit|wi(g |nc|nw)|wmlb|wonu|x700|yas\-|your|zeto|zte\-/i.test(a.substr(0,4)))isMobile = true})(navigator.userAgent||navigator.vendor||window.opera);
            return isMobile;
        }

        static setStyleDisplay(element, showHide) {
            if (showHide == true) {
                element.style.display = "";
            } else {
                element.style.display = "none";
            }
        }

    }

    class Base64 {
        static get code() { return 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/='; }

        constructor() {
        }

        /**
         * Encode string into Base64, as defined by RFC 4648 [http://tools.ietf.org/html/rfc4648]
         * (instance method extending String object). As per RFC 4648, no newlines are added.
         *
         * @param {String} str The string to be encoded as base-64
         * @param {Boolean} [utf8encode=false] Flag to indicate whether str is Unicode string to be encoded
         *   to UTF8 before conversion to base64; otherwise string is assumed to be 8-bit characters
         * @returns {String} Base64-encoded string
         */
        static encode(str, utf8encode) {
            utf8encode =  (typeof utf8encode == 'undefined') ? false : utf8encode;
            let o1, o2, o3, bits, h1, h2, h3, h4, e=[], pad = '', c, plain, coded;
            let b64 = Base64.code;

            plain = utf8encode ? Utf8.encode(str) : str;

            c = plain.length % 3;  // pad string to length of multiple of 3
            if (c > 0) { while (c++ < 3) { pad += '='; plain += '\0'; } }
            // note: doing padding here saves us doing special-case packing for trailing 1 or 2 chars

            for (c=0; c<plain.length; c+=3) {  // pack three octets into four hexets
                o1 = plain.charCodeAt(c);
                o2 = plain.charCodeAt(c+1);
                o3 = plain.charCodeAt(c+2);

                bits = o1<<16 | o2<<8 | o3;

                h1 = bits>>18 & 0x3f;
                h2 = bits>>12 & 0x3f;
                h3 = bits>>6 & 0x3f;
                h4 = bits & 0x3f;

                // use hextets to index into code string
                e[c/3] = b64.charAt(h1) + b64.charAt(h2) + b64.charAt(h3) + b64.charAt(h4);
            }
            coded = e.join('');  // join() is far faster than repeated string concatenation in IE

            // replace 'A's from padded nulls with '='s
            coded = coded.slice(0, coded.length-pad.length) + pad;

            return coded;
        }

        static decode(str, utf8decode) {
            utf8decode =  (typeof utf8decode == 'undefined') ? false : utf8decode;
            let o1, o2, o3, h1, h2, h3, h4, bits, d=[], plain, coded;
            let b64 = Base64.code;

            coded = utf8decode ? Utf8.decode(str) : str;

            for (let c=0; c<coded.length; c+=4) {  // unpack four hexets into three octets
                h1 = b64.indexOf(coded.charAt(c));
                h2 = b64.indexOf(coded.charAt(c+1));
                h3 = b64.indexOf(coded.charAt(c+2));
                h4 = b64.indexOf(coded.charAt(c+3));

                bits = h1<<18 | h2<<12 | h3<<6 | h4;

                o1 = bits>>>16 & 0xff;
                o2 = bits>>>8 & 0xff;
                o3 = bits & 0xff;

                d[c/4] = String.fromCharCode(o1, o2, o3);
                // check for padding
                if (h4 == 0x40) d[c/4] = String.fromCharCode(o1, o2);
                if (h3 == 0x40) d[c/4] = String.fromCharCode(o1);
            }
            plain = d.join('');  // join() is far faster than repeated string concatenation in IE

            return utf8decode ? Utf8.decode(plain) : plain;
        }
    }


    const kVisitorRetransmitPeriod = 100000;
    const kChatRetransmitPeriod = 5000;

    class TestClient {

        constructor(fragmentID, webSocket) {

            this._fragmentID = fragmentID;
            this._wsp = webSocket;
            this._pageBodyElement = document.getElementsByTagName('body')[0];
            this._pageHeadElement = document.getElementsByTagName('head')[0];

            this._session = new Session(this._fragmentID, this._wsp,
                document, navigator, screen, kVisitorRetransmitPeriod, this._pageHeadElement);
        }

        testRequest() {

            let parser = new UAParser(); //script/UAParser.js
            let uastring = navigator.userAgent;

            parser.setUA(uastring);

            const payload = {
                'fragmentID': 	'4879cb8b-789b-4609-813c-0b2278edc312',
                    'visitorID':   '1234', //JSON.parse(TC.cookieStorage.read(inFragmentId)).visitorID,
                    'entryUrl': 	'', // inSessionData.entryUrl,
                    'referrerUrl': 	'' //inSessionData.refUrl
            };

            let request = {
                'oty' : 'tco',
                'mty' : 'request',
                'rty' : 'session/new',
                package: payload
            };

            let response;
            this._wsp.open()
                .then(() => { // actually sends {foo: 'bar', id: 'xxx'}
                    this._wsp.sendRequest({message: request})
                        .then( fulfill => {
                            console.log('fulfill request:', fulfill.message);
                        })
                        .catch((error) => {
                            console.log('catch error:', error);
                        })
                        .finally(() => {
                        console.log('finally');
                    });
                    console.log('message sent:' , request);
                });

        }
    }

    // INIT
    const kFragmentID = '4d546d07-1fa1-4b25-8b2a-d6acb0d8e72d';
    const kWebsocketServerURL = 'ws://localhost:7080';
    window.ETSChat = ETSChat;
    window.ETSChat.fragmentID = kFragmentID;

    // storage
    window.ETSChat.localStorage = new LocalStorage();
    window.ETSChat.sessionStorage = new SessionStorage();
    window.ETSChat.cookieStorage = new CookieStorage();

    let bodyElement0 = document.getElementsByTagName('body')[0];
    let headElement0 = document.getElementsByTagName('head')[0];

    window.ETSChat.alignmentManager = new AlignmentManager(bodyElement0, headElement0);

    window.ETSChat.socket = new WebSocketAsPromised(kWebsocketServerURL, {
        packMessage: data => JSON.stringify(data),
        unpackMessage: message => JSON.parse(message),
        attachRequestId: (data, requestId) => Object.assign({id: requestId}, data), // attach requestId to message as `id` field
        extractRequestId: data => data && data.id,                                  // read requestId from message `id` field
    });

    window.ETSChat.socket.onMessage.addListener(message => {
        // response = message;
        // console.log('ETSChat.socket.onMessage => ', message);
        let messageObject = JSON.parse(message);
        if(messageObject.message !== null && messageObject.message !== undefined) {
            if(messageObject.message.rty !== null && messageObject.message.rty !== undefined) {
                if(messageObject.message.rty === 'chat/message') {
                    if(messageObject.message.package !== null && messageObject.message.package !== undefined) {
                        if(messageObject.message.package.visitorID !== null && messageObject.message.package.visitorID !== undefined) {
                            let me = ETSChat.session._visitor._id;
                            if(messageObject.message.package.visitorID !== me) {
                                ETSChat.chatWindowController.addMessage(messageObject.message);
                            }
                        }
                    }
                }
            }
        }
    });

    window.ETSChat.chatButtonView = new ChatButtonView(ETSChat.alignmentManager, bodyElement0);
    window.ETSChat.chatButtonController = new ChatButtonController(ETSChat.chatButtonView, null ,kFragmentID);
    window.ETSChat.chatWindowView = new ChatWindowView(0, bodyElement0, headElement0 );
    window.ETSChat.chatWindowController = new ChatWindowController(window.ETSChat.socket, 100000, kFragmentID, ETSChat.chatWindowView)

    window.testClient = new TestClient(kFragmentID, window.ETSChat.socket);
    //window.testClient.testRequest();
    console.log("done with load");
})(window);
